#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   model.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Model types.
"""

# pylint: disable=R0902

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional

import numpy as np

from datature.nexus.cli.deploy.local.types.skeleton import Skeleton
from datature.nexus.cli.deploy.local.utils.inference import (
    PredictionTaskType,
    TensorDescription,
)
from datature.nexus.cli.deploy.local.utils.label_map import LabelMap


class ModelBackend(Enum):
    """Backend used for model inference."""

    ONNX = "ONNX"
    TENSORFLOW = "TensorFlow"
    PYTORCH = "PyTorch"
    TFLITE = "TFLite"
    OPENVINO = "OpenVINO"


class ModelType(Enum):
    """Model architecture enum."""

    ULTRALYTICS = "Ultralytics"
    POLYBIUS_V1 = "PolybiusV1"
    OBJECT_DETECTION = "ObjectDetection"
    STREAMING = "STREAMING"


class ModelArchitecture(Enum):
    """Model architecture enum."""

    DEEPLABV3 = "DEEPLABV3"
    YOLOV4 = "YOLOV4"
    YOLOX = "YOLOX"
    PALIGEMMA = "PALIGEMMA"


@dataclass
class Model:
    """Base model class."""

    model: Any

    versions: List[int]
    backend: ModelBackend
    model_type: ModelType

    label_map: LabelMap

    inputs: List[TensorDescription]
    outputs: List[TensorDescription]

    task_type: PredictionTaskType


@dataclass
class KeypointModel(Model):
    """Keypoint detection model class."""

    skeletons: Optional[Dict[str, Skeleton]] = None


@dataclass
class VideoClassificationModel(Model):
    """Video classification model class."""

    memory: Dict[str, np.ndarray] = field(default_factory=dict)
    memory_initialized: bool = False



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   onnx_loader.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   ONNX Loader class.
"""

# pylint: disable=R0914

import ast
import json
import logging
import queue
import subprocess
import sys
from typing import Dict

try:
    import onnxruntime as ort

except (ModuleNotFoundError, ImportError):
    print("Installing onnxruntime...")
    subprocess.check_call([sys.executable, "-m", "pip", "install", "-q", "onnxruntime"])

    try:
        subprocess.check_call(
            [sys.executable, "-m", "pip", "install", "-q", "onnxruntime-gpu"]
        )
    except subprocess.CalledProcessError:
        pass

    import onnxruntime as ort

from datature.nexus.cli.deploy.local.loaders.base_loader import BaseLoader
from datature.nexus.cli.deploy.local.types.skeleton import Skeleton
from datature.nexus.cli.deploy.local.utils.colors import get_color_for_label_id
from datature.nexus.cli.deploy.local.utils.common import start_spinner
from datature.nexus.cli.deploy.local.utils.inference import (
    PredictionTaskType,
    TensorDescription,
    TensorDtype,
)
from datature.nexus.cli.deploy.local.utils.label_map import (
    LabelMap,
    LabelMapEntry,
    LabelMapFromPbtxt,
)
from datature.nexus.cli.deploy.local.utils.model import (
    KeypointModel,
    Model,
    ModelArchitecture,
    ModelBackend,
    ModelType,
)

logger = logging.getLogger("datature-nexus")

ULTRALYTICS_MODELS = ["yolov8", "yolov9", "yolo11"]

ULTRALYTICS_TASKS = {
    "ObjectDetection": "detect",
    "InstanceSegmentation": "segment",
    "SemanticSegmentation": "segment",
    "KeypointDetection": "pose",
    "ImageClassification": "classify",
}

POLYBIUS_PREDICTION_TYPES: Dict[str, PredictionTaskType] = {
    "bbox": PredictionTaskType.OBJECT_DETECTION,
    "instance": PredictionTaskType.INSTANCE_SEGMENTATION,
    "semantic": PredictionTaskType.SEMANTIC_SEGMENTATION,
    "semantic_3d": PredictionTaskType.SEMANTIC_3D_SEGMENTATION,
    "classification": PredictionTaskType.IMAGE_CLASSIFICATION,
    "keypoints": PredictionTaskType.KEYPOINT_DETECTION,
}

LEGACY_MODEL_ARCHITECTURES: Dict[str, ModelArchitecture] = {
    "deeplabv3": ModelArchitecture.DEEPLABV3,
    "yolov4": ModelArchitecture.YOLOV4,
    "yolox": ModelArchitecture.YOLOX,
}

LEGACY_MODEL_TASK_TYPES: Dict[str, PredictionTaskType] = {
    "deeplabv3": PredictionTaskType.SEMANTIC_SEGMENTATION,
    "yolov4": PredictionTaskType.OBJECT_DETECTION,
    "yolox": PredictionTaskType.OBJECT_DETECTION,
}

EXECUTION_PROVIDERS = {
    "CPU": "CPUExecutionProvider",
    "CUDA": "CUDAExecutionProvider",
}

ONNX_SUPPORTED_DTYPES: Dict[str, TensorDtype] = {
    "tensor(float)": TensorDtype.FP32,
    "tensor(float16)": TensorDtype.FP16,
    "tensor(uint8)": TensorDtype.UINT8,
    "tensor(uint16)": TensorDtype.UINT16,
    "tensor(uint32)": TensorDtype.UINT32,
    "tensor(uint64)": TensorDtype.UINT64,
    "tensor(int8)": TensorDtype.INT8,
    "tensor(int16)": TensorDtype.INT16,
    "tensor(int32)": TensorDtype.INT32,
    "tensor(int64)": TensorDtype.INT64,
    "tensor(bool)": TensorDtype.BOOL,
}


def _tensor_description_from_onnx_node_arg(node_arg: ort.NodeArg):
    """Create TensorDescription from ONNX NodeArg."""
    if node_arg.type not in ONNX_SUPPORTED_DTYPES:
        raise ValueError(
            f"Cannot infer TensorDescription from ONNX NodeArg of type "
            f"{node_arg.type}, expected one of: "
            f"{', '.join(ONNX_SUPPORTED_DTYPES.keys())}"
        )

    name = node_arg.name

    # Important! We don't preserve any distinctions between the unknown
    # tensor dimension sizes, so we assume they can vary freely.
    shape = tuple(map(lambda size: -1 if "unk" in str(size) else size, node_arg.shape))

    return TensorDescription(
        name=name, dtype=ONNX_SUPPORTED_DTYPES[node_arg.type], shape=shape
    )


class ONNXLoader(BaseLoader):
    """ONNX Loader class."""

    @start_spinner
    def load(self, message_queue: queue.Queue = queue.Queue()):
        """Load ONNX model."""
        message_queue.put("Loading ONNX model...")

        session_options = ort.SessionOptions()
        session_options.log_severity_level = 3
        session = ort.InferenceSession(
            self._model_path,
            sess_options=session_options,
            providers=[self._execution_provider],
        )

        model_type = session.get_modelmeta().description

        if any(model_name in self._model_path for model_name in ULTRALYTICS_MODELS):
            self._model = self._load_ultralytics_yolo_model(session)
        elif "Polybius" in model_type:
            self._model = self._load_polybius_v1_model(session)
        elif "object_detection" in model_type:
            self._model = self._load_object_detection_model(session)
        else:
            self._model = self._load_legacy_model(session)

        return self._model, "ONNX model loaded."

    def _load_ultralytics_yolo_model(self, session: ort.InferenceSession) -> Model:
        """Load Ultralytics YOLO model."""
        inputs = session.get_inputs()
        outputs = session.get_outputs()

        model_description = session.get_modelmeta().description
        metadata_map = session.get_modelmeta().custom_metadata_map
        label_map_entries = []

        if "colors" in metadata_map:
            color_map = ast.literal_eval(metadata_map["colors"])
        else:
            color_map = None

        label_information = ast.literal_eval(metadata_map["names"])

        if isinstance(label_information, list):
            # For compatibility with both new and old versions of yolov8/9
            label_information = dict(enumerate(label_information))

        for label_id, label_name in label_information.items():
            if color_map is None or label_name not in color_map:
                color = get_color_for_label_id(label_id)
            else:
                color = color_map[label_name]

            label_map_entries.append(
                LabelMapEntry(
                    label_id=label_id,
                    name=label_name,
                    color=color,
                )
            )

        skeletons: Dict[str, Skeleton] = {}

        if "cls" in model_description:
            task_type = PredictionTaskType.IMAGE_CLASSIFICATION

        elif "seg" in model_description:
            task_type = PredictionTaskType.INSTANCE_SEGMENTATION

        elif "pose" in model_description:
            task_type = PredictionTaskType.KEYPOINT_DETECTION
            skeletons_json = ast.literal_eval(metadata_map["skeletons"])["skeletons"]
            for sk_name, sk_object in skeletons_json.items():
                if "name" not in sk_object:
                    sk_object["name"] = sk_name
                skeletons[sk_name] = Skeleton.from_json(sk_object)

            return KeypointModel(
                model=session,
                versions=[1],
                backend=ModelBackend.ONNX,
                task_type=task_type,
                model_type=ModelType.ULTRALYTICS,
                label_map=LabelMap(label_map_entries),
                inputs=list(map(_tensor_description_from_onnx_node_arg, inputs)),
                outputs=list(map(_tensor_description_from_onnx_node_arg, outputs)),
                skeletons=skeletons,
            )

        else:
            task_type = PredictionTaskType.OBJECT_DETECTION

        return Model(
            model=session,
            versions=[1],
            backend=ModelBackend.ONNX,
            task_type=task_type,
            model_type=ModelType.ULTRALYTICS,
            label_map=LabelMap(label_map_entries),
            inputs=list(map(_tensor_description_from_onnx_node_arg, inputs)),
            outputs=list(map(_tensor_description_from_onnx_node_arg, outputs)),
        )

    def _load_polybius_v1_model(self, session: ort.InferenceSession) -> Model:
        """Load Polybius v1 model."""
        inputs = session.get_inputs()
        outputs = session.get_outputs()

        if len(inputs) != 1:
            raise RuntimeError(
                f"Cannot load Polybius model having {len(inputs)} " "inputs, expected 1"
            )

        model_description = session.get_modelmeta().description

        model_version, prediction_type = [
            info.lower() for info in model_description.split("_", 1)[-1].split(".")
        ]

        if model_version != "v1":
            raise RuntimeError(
                f"Cannot load Polybius model with version {model_version}, "
                "expected 'v1'"
            )

        metadata_map = session.get_modelmeta().custom_metadata_map

        if "colors" in metadata_map:
            color_map = json.loads(metadata_map["colors"])
        else:
            color_map = None

        with open(self._label_map_path, "r", encoding="utf-8") as label_map_file:
            label_map = LabelMapFromPbtxt(label_map_file, color_map)

        return Model(
            model=session,
            versions=[1],
            backend=ModelBackend.ONNX,
            task_type=POLYBIUS_PREDICTION_TYPES[prediction_type],
            model_type=ModelType.POLYBIUS_V1,
            label_map=label_map,
            inputs=list(map(_tensor_description_from_onnx_node_arg, inputs)),
            outputs=list(map(_tensor_description_from_onnx_node_arg, outputs)),
        )

    def _load_object_detection_model(self, session: ort.InferenceSession) -> Model:
        """Load Tensorflow Object Detection model."""
        inputs = session.get_inputs()
        outputs = session.get_outputs()

        if len(inputs) != 1:
            raise RuntimeError(
                f"Cannot load Object Detection model having {len(inputs)} "
                "inputs, expected 1"
            )

        task_type = PredictionTaskType.OBJECT_DETECTION

        for output in outputs:
            if "masks" in output.name:
                task_type = PredictionTaskType.INSTANCE_SEGMENTATION
                break

        metadata_map = session.get_modelmeta().custom_metadata_map

        if "colors" in metadata_map:
            color_map = json.loads(metadata_map["colors"])
        else:
            color_map = None

        with open(self._label_map_path, "r", encoding="utf-8") as label_map_file:
            label_map = LabelMapFromPbtxt(label_map_file, color_map)

        return Model(
            model=session,
            versions=[1],
            backend=ModelBackend.ONNX,
            task_type=task_type,
            model_type=ModelType.OBJECT_DETECTION,
            label_map=label_map,
            inputs=list(map(_tensor_description_from_onnx_node_arg, inputs)),
            outputs=list(map(_tensor_description_from_onnx_node_arg, outputs)),
        )

    def _load_legacy_model(self, session: ort.InferenceSession) -> Model:
        """Load legacy model."""
        inputs = session.get_inputs()
        outputs = session.get_outputs()
        model_type = session.get_modelmeta().description

        if len(inputs) != 1:
            raise RuntimeError(
                f"Cannot load {model_type} model having {len(inputs)} "
                "inputs, expected 1"
            )

        if (
            model_type not in LEGACY_MODEL_ARCHITECTURES
            or model_type not in LEGACY_MODEL_TASK_TYPES
        ):
            raise RuntimeError(
                f"Cannot load {model_type} as legacy model, we have no "
                "knowledge of this model type"
            )

        metadata_map = session.get_modelmeta().custom_metadata_map
        if "colors" in metadata_map:
            color_map = json.loads(metadata_map["colors"])
        else:
            color_map = None

        with open(self._label_map_path, "r", encoding="utf-8") as label_map_file:
            label_map = LabelMapFromPbtxt(label_map_file, color_map)

        return Model(
            model=session,
            versions=[1],
            backend=ModelBackend.ONNX,
            model_type=LEGACY_MODEL_ARCHITECTURES[model_type],
            label_map=label_map,
            inputs=list(map(_tensor_description_from_onnx_node_arg, inputs)),
            outputs=list(map(_tensor_description_from_onnx_node_arg, outputs)),
            task_type=LEGACY_MODEL_TASK_TYPES[model_type],
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██
@File    :   logger.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Logger configuration for the local deployment server.
"""

from io import StringIO

LOG_STREAM = StringIO()

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "()": "uvicorn.logging.DefaultFormatter",
            "fmt": "%(asctime)s.%(msecs)03d\t%(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
            "use_colors": None,
        },
        "access": {
            "()": "uvicorn.logging.AccessFormatter",
            "fmt": "%(asctime)s.%(msecs)03d\t%(client_addr)s - %(request_line)s %(status_code)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
            "use_colors": None,
        },
    },
    "handlers": {
        "default": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": LOG_STREAM,
        },
        "default_stdout": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
        "access": {
            "formatter": "access",
            "class": "logging.StreamHandler",
            "stream": LOG_STREAM,
        },
    },
    "loggers": {
        "uvicorn": {
            "handlers": ["default", "default_stdout"],
            "level": "INFO",
            "propagate": False,
        },
        "uvicorn.error": {"level": "INFO"},
        "uvicorn.access": {
            "handlers": ["access"],
            "level": "INFO",
            "propagate": False,
        },
        "litserve": {
            "handlers": ["default", "default_stdout"],
            "level": "ERROR",
            "propagate": False,
        },
        "litserve.server": {
            "handlers": ["access"],
            "level": "ERROR",
            "propagate": False,
        },
    },
}


class LogStreamTracker:
    """Tracks the position in the log stream across server lifetime."""

    pointer = 0



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   upload_session.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Asset Upload Session
"""
# pylint: disable=R1732,R0902,R0912,R0913,R0914,R0915,R0917,E0203,W0201,W0718

import concurrent.futures
import logging
import os
import struct
import tempfile
import threading
import time
from contextlib import ContextDecorator
from typing import List, Optional, Tuple, Union

import crc32c
import cv2
import requests
from filetype import filetype

from datature.nexus import config, error
from datature.nexus.api.asset.multipart import MultipartHandler
from datature.nexus.api.operation import Operation
from datature.nexus.api.types import AssetFilePart, OperationStatusOverview
from datature.nexus.client_context import ClientContext, RestContext
from datature.nexus.models import (
    CancelResponse,
    MultipartAbortResponse,
    MultipartCompleteResponse,
    MultipartPartStatus,
)
from datature.nexus.models import MultipartUploadSession as MultipartUploadSessionModel
from datature.nexus.models import MultipartUploadSignedUrl
from datature.nexus.models import UploadSession as UploadSessionModel
from datature.nexus.models import UploadSessionAssetItem
from datature.nexus.utils import file_signature, utils

logger = logging.getLogger("datature-nexus")


class UploadSession(RestContext, ContextDecorator):
    """Datature Asset Upload Session Class.

    :param client_context: An instance of ClientContext.
    :param groups: A list of group names to categorize the upload. Default is None.
    :param background:
        A flag indicating whether the upload should run in the background. Default is False.
    """

    def __init__(
        self,
        client_context: ClientContext,
        groups: Optional[List[str]] = None,
        background: bool = False,
    ):
        """Initialize the API Resource."""
        super().__init__(client_context)
        self._local = threading.local()

        self._operation = None  # Lazy initialization if needed
        self.assets = []
        self.file_name_map = {}
        self.upload_session_ids = []
        self.operation_ids = []

        self.groups = groups if groups is not None else ["main"]
        self.background = background

        self.abort_event = threading.Event()

    @property
    def operation(self):
        """Initialize operation."""
        if self._operation is None:
            self._operation = Operation(self._context)
        return self._operation

    def _init_http_session(self, abort_events: Optional[List[threading.Event]] = None):
        """Initialize local session and retry policy."""
        if abort_events is None:
            abort_events = []

        self._local.abort_event = threading.Event()
        abort_events.append(self._local.abort_event)

        self._local.http_session = utils.init_gcs_upload_session(abort_events)

    def __enter__(self):
        return self

    def __exit__(self, _exc_type, exc_val, _exc_tb):
        """Exit function.
        The function will be called if an exception is raised inside the context manager
        """
        if exc_val is not None:
            logger.warning("Upload session error: %s", exc_val)
            self._cancel_upload_session()
            raise error.Error(exc_val)

        # check asset length
        if len(self.assets) == 0 and len(self.operation_ids) == 0:
            raise error.Error("Assets to upload is empty")

        # call API to get signed url
        if self.assets:
            response = self._upload_assets()
            self.operation_ids.append(response.op_id)

        if self.background:
            return {"op_ids": self.operation_ids}

        # Wait server finish generate thumbnail
        self.wait_until_done()

        return {"op_ids": self.operation_ids}

    def __len__(self):
        """Over write len function."""
        return len(self.file_name_map)

    def add_path(self, file_path: str):
        """
        Add asset to upload.

        :param file_path: The path of the file to upload.
        """
        if not os.path.exists(file_path):
            raise error.Error("Cannot find the Asset file")

        if os.path.isdir(file_path):
            file_paths = utils.find_all_assets(file_path)
        else:
            file_paths = [file_path]

        for each_file_path in file_paths:
            self._generate_metadata(os.path.basename(each_file_path), each_file_path)
            # check current asset size
            self._check_current_asset_size()

    def add_bytes(
        self,
        file_bytes: bytes,
        filename: str,
    ):
        """Attach file in bytes to upload session

        :param file_bytes: The bytes of the file to upload.
        :param filename: The name of the file to upload, should include the file extension.
        """
        file_mime_type = file_signature.get_file_mime_by_signature(file_bytes)

        if file_mime_type is None:
            raise TypeError(f"Unsupported file: {filename}")

        # Create a temporary directory
        temp_dir = tempfile.mkdtemp()

        # Create a temporary file path
        temp_file_path = os.path.join(temp_dir, filename)

        # Write file bytes to the temporary file
        with open(temp_file_path, "wb") as temp_file:
            temp_file.write(file_bytes)

        self._generate_metadata(os.path.basename(temp_file_path), temp_file_path)
        # check current asset size
        self._check_current_asset_size()

    def _generate_metadata(self, filename: str, file_path: str):
        """process the file to asset metadata."""
        size = os.path.getsize(file_path)

        if size < config.ASSET_UPLOAD_SESSION_MULTIPART_MIN_SIZE:
            file_hash = crc32c.CRC32CHash()

            with open(file_path, "rb") as file:
                chunk = file.read(config.FILE_CHUNK_SIZE)
                while chunk:
                    file_hash.update(chunk)
                    chunk = file.read(config.FILE_CHUNK_SIZE)

            # To fix the wrong crc32 caused by mac M1 clip
            crc32 = struct.unpack(">l", file_hash.digest())[0]

        else:
            crc32 = 0

        guess_result = filetype.guess(file_path)
        mime = utils.ASSET_FILE_EXTENSION_TO_MIME_TYPE_MAP.get(
            os.path.splitext(file_path)[1],
            guess_result.mime if guess_result else None,
        )

        if self.file_name_map.get(filename) is not None:
            raise error.Error(
                f"Cannot add multiple files with the same name, {filename}"
            )

        if filename and size and mime:
            if mime in utils.SUPPORTED_VIDEO_MIME_TYPES:
                if size > config.VIDEO_MAX_SIZE:
                    raise error.Error(
                        f"Video {filename} size exceeds the limit: "
                        f"{config.VIDEO_MAX_SIZE / 1024 / 1024 / 1024} GB"
                    )

                frames = 1
                try:
                    cap = cv2.VideoCapture(file_path)
                    if not cap.isOpened():
                        raise error.Error(f"Failed to open video file: {filename}")

                    frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                    if frames <= 0:
                        raise error.Error(f"Invalid frame count for video: {filename}")

                    cap.release()

                except Exception as exc:
                    logger.warning(
                        "Error reading video file %s: %s, "
                        "Video will still be attempted to be uploaded",
                        filename,
                        str(exc),
                    )

                asset_metadata = {
                    "filename": filename,
                    "size": size,
                    "crc32c": crc32,
                    "mime": mime,
                    "frames": frames,
                    "encoder": {"profile": "h264Saver", "everyNthFrame": 1},
                }

            elif mime in utils.SUPPORTED_IMAGE_MIME_TYPES:
                if size > config.IMAGE_MAX_SIZE:
                    raise error.Error(
                        f"Image {filename} size exceeds the limit: "
                        f"{config.IMAGE_MAX_SIZE / 1024 / 1024} MB"
                    )

                asset_metadata = {
                    "filename": filename,
                    "size": size,
                    "crc32c": crc32,
                    "mime": mime,
                }

            elif mime in utils.SUPPORTED_MEDICAL_3D_MIME_TYPES:
                if size > config.MEDICAL_3D_MAX_SIZE:
                    raise error.Error(
                        f"Medical 3D {filename} size exceeds the limit: "
                        f"{config.MEDICAL_3D_MAX_SIZE / 1024 / 1024 / 1024} GB"
                    )

                asset_metadata = {
                    "filename": filename,
                    "size": size,
                    "crc32c": crc32,
                    "mime": mime,
                }

            else:
                raise error.Error(
                    f"Asset MIME type {mime} is not supported. "
                    "Supported MIME types: "
                    + ", ".join(utils.SUPPORTED_IMAGE_MIME_TYPES)
                    + ", "
                    + ", ".join(utils.SUPPORTED_VIDEO_MIME_TYPES)
                    + ", "
                    + ", ".join(utils.SUPPORTED_MEDICAL_3D_MIME_TYPES)
                )

            self.assets.append(asset_metadata)
            self.file_name_map[filename] = {"path": file_path}

            logger.debug("Add asset: %s", asset_metadata)
        else:
            raise error.Error("Unsupported asset file")

    def _upload_file_through_signed_url(
        self, asset_upload, abort_event: threading.Event
    ) -> Tuple[str, bool]:
        """
        Upload a file through signed url.

        :param asset_upload: The asset upload response containing metadata
        :param abort_event: Event to signal abort
        :return: A tuple of the filename and a boolean indicating if the upload was successful
        """
        filename = asset_upload.get("metadata").get("filename")
        file_path = self.file_name_map.get(filename)["path"]

        # Check for abort before proceeding
        if abort_event.is_set():
            return filename, False

        try:
            # upload asset to GCP
            with open(file_path, "rb") as file:
                self._local.http_session.request(
                    asset_upload.upload.method,
                    asset_upload.upload.url,
                    headers=asset_upload.upload.headers,
                    data=file,
                    timeout=config.REQUEST_TIME_OUT_SECONDS,
                )
            return filename, True

        except requests.exceptions.RequestException:
            if abort_event.is_set():
                return filename, False
            raise

    def _upload_part(
        self,
        multipart_handler: MultipartHandler,
        multipart_upload_session_response: MultipartUploadSessionModel,
        part: AssetFilePart,
        asset_upload: UploadSessionAssetItem,
        file_abort_event: threading.Event,
        abort_event: threading.Event,
    ):
        """
        Upload a single part of a large asset file.

        :param multipart_handler: The MultipartHandler instance managing the file parts.
        :param multipart_upload_session_response: The response from the multipart upload session.
        :param part: The part of the file to upload.
        :param asset_upload: The asset upload metadata.
        :param file_abort_event: Event to signal abort for this file's parts.
        :param abort_event: Global event to signal abort (e.g. from keyboard interrupt).
        :return: None
        """
        retry_count = 0

        while retry_count < config.ASSET_UPLOAD_MULTIPART_MAX_RETRIES:
            try:
                # Check both abort events before proceeding
                if file_abort_event.is_set() or abort_event.is_set():
                    raise error.Error("Upload interrupted by user")

                part_url_response = self.requester.PUT(
                    f"/projects/{self.project_id}/multipartassetuploads/"
                    f"{multipart_upload_session_response.id}/parts/{part.part_number}",
                    response_type=MultipartUploadSignedUrl,
                )

                part_data = multipart_handler.read_part_data(part)

                # Upload part to GCS
                upload_response = self._local.http_session.request(
                    part_url_response.method,
                    part_url_response.url,
                    headers=part_url_response.headers,
                    data=part_data,
                    timeout=config.REQUEST_TIME_OUT_SECONDS,
                )

                if not upload_response.ok:
                    raise error.Error(
                        f"Part {part.part_number} upload failed: "
                        f"{upload_response.status_code} {upload_response.text}"
                    )

                upload_response_headers = {
                    k.lower(): v for k, v in upload_response.headers.items()
                }

                gcs_upload_response_data = {
                    "responseHeaders": {
                        header: upload_response_headers.get(header.lower(), "")
                        for header in part_url_response.response_headers
                    },
                }

                if part_url_response.response_body:
                    gcs_upload_response_data["responseBody"] = part_url_response.text

                part_status_response = self.requester.POST(
                    f"/projects/{self.project_id}/multipartassetuploads/"
                    f"{multipart_upload_session_response.id}/parts/"
                    f"{part.part_number}/complete",
                    request_body=gcs_upload_response_data,
                    response_type=MultipartPartStatus,
                )

                if not part_status_response.completed:
                    raise error.Error(
                        f"Part {part.part_number} upload completion "
                        f"not registered by server for {asset_upload.metadata.filename}"
                    )

                logger.debug(
                    "Uploaded part %d/%d for %s",
                    part.part_number,
                    multipart_upload_session_response.part_count,
                    asset_upload.metadata.filename,
                )

                return

            except Exception as exc:
                if file_abort_event.is_set() or abort_event.is_set():
                    raise

                retry_count += 1
                if retry_count >= config.ASSET_UPLOAD_MULTIPART_MAX_RETRIES:
                    logger.error(
                        "Failed to upload part %d for %s after %d retries: %s",
                        part.part_number,
                        asset_upload.metadata.filename,
                        config.ASSET_UPLOAD_MULTIPART_MAX_RETRIES,
                        str(exc),
                    )

                    raise error.Error(
                        f"Failed to upload part {part.part_number} for "
                        f"{asset_upload.metadata.filename} after "
                        f"{config.ASSET_UPLOAD_MULTIPART_MAX_RETRIES} retries: {str(exc)}"
                    )

                logger.warning(
                    "Error: %s, Retrying part %d upload for %s (attempt %d/%d)",
                    str(exc),
                    part.part_number,
                    asset_upload.metadata.filename,
                    retry_count + 1,
                    config.ASSET_UPLOAD_MULTIPART_MAX_RETRIES,
                )

                time.sleep(2**retry_count)

    def _upload_multipart_file_through_signed_url(
        self,
        multipart_executor: concurrent.futures.ThreadPoolExecutor,
        upload_session_id: str,
        asset_upload: UploadSessionAssetItem,
        abort_event: threading.Event,
    ) -> Tuple[str, bool]:
        """
        Upload a large file to GCS using multipart upload.

        :param multipart_executor: The thread pool executor for multipart uploads.
        :param upload_session_id: The upload session ID
        :param asset_upload: The asset upload response containing metadata
        :param abort_event: Global event to signal abort (e.g. from keyboard interrupt)
        :return: A tuple of the filename and a boolean indicating if the upload was successful
        """
        file_path = self.file_name_map.get(asset_upload.metadata.filename, {}).get(
            "path"
        )

        logger.debug(
            "Starting multipart upload for %s (%.2f MB)",
            asset_upload.metadata.filename,
            asset_upload.metadata.size / (1024 * 1024),
        )

        part_abort_event = threading.Event()

        if abort_event.is_set():
            raise error.Error("Upload interrupted by user")

        self._init_http_session([abort_event, part_abort_event])

        multipart_upload_session_response = self.requester.POST(
            f"/projects/{self.project_id}/assetuploadsessions/{upload_session_id}/multipart",
            request_body={
                "filename": asset_upload.metadata.filename,
                "mime": asset_upload.metadata.mime,
                "size": asset_upload.metadata.size,
            },
            response_type=MultipartUploadSessionModel,
        )

        try:
            multipart_handler = MultipartHandler(
                file_path, multipart_upload_session_response.part_count
            )

            logger.debug(
                "File %s split into %d parts",
                asset_upload.metadata.filename,
                multipart_upload_session_response.part_count,
            )

            futures = []
            for part in multipart_handler.parts:
                futures.append(
                    multipart_executor.submit(
                        self._upload_part,
                        multipart_handler,
                        multipart_upload_session_response,
                        part,
                        asset_upload,
                        part_abort_event,
                        abort_event,
                    )
                )

            done, _ = concurrent.futures.wait(
                futures, return_when=concurrent.futures.FIRST_EXCEPTION
            )

            for future in done:
                exc = future.exception()
                if exc is not None:
                    part_abort_event.set()
                    raise error.Error("Upload interrupted by user") from exc

            complete_response = self.requester.POST(
                f"/projects/{self.project_id}/multipartassetuploads/"
                f"{multipart_upload_session_response.id}/complete",
                response_type=MultipartCompleteResponse,
            )

            logger.debug(
                "Completed multipart upload for %s: %s",
                asset_upload.metadata.filename,
                complete_response,
            )

            return asset_upload.metadata.filename, True

        except Exception as exc:
            logger.debug(
                "Error: %s, Aborting multipart upload for %s",
                str(exc),
                asset_upload.metadata.filename,
            )

            self.requester.DELETE(
                f"/projects/{self.project_id}/multipartassetuploads/"
                f"{multipart_upload_session_response.id}",
                response_type=MultipartAbortResponse,
                ignore_errno=[404],
            )

            return asset_upload.metadata.filename, False

    def _upload_assets(self):
        """Use ThreadPoolExecutor to upload asset files to GCS."""
        upload_session_response = self.requester.POST(
            f"/projects/{self.project_id}/assetuploadsessions",
            request_body={"groups": self.groups, "assets": self.assets},
            response_type=UploadSessionModel,
        )
        self.upload_session_ids.append(upload_session_response.id)

        large_assets = []
        small_assets = []

        for asset, asset_response in zip(self.assets, upload_session_response.assets):
            if asset["size"] > config.ASSET_UPLOAD_SESSION_MULTIPART_MIN_SIZE:
                large_assets.append((asset, asset_response))
            else:
                small_assets.append((asset, asset_response))

        cpu_count = int(os.cpu_count() or 1)
        multipart_executor = None
        upload_executor = None
        futures = []

        try:
            if large_assets:
                multipart_workers = max(
                    int(cpu_count * config.ASSET_UPLOAD_SESSION_WORKERS_RATIO / 2), 1
                )
                normal_workers = max(
                    int(cpu_count * config.ASSET_UPLOAD_SESSION_WORKERS_RATIO / 2), 1
                )
                multipart_executor = concurrent.futures.ThreadPoolExecutor(
                    initializer=self._init_http_session, max_workers=multipart_workers
                )
            else:
                normal_workers = max(
                    int(cpu_count * config.ASSET_UPLOAD_SESSION_WORKERS_RATIO), 1
                )

            upload_executor = concurrent.futures.ThreadPoolExecutor(
                initializer=self._init_http_session, max_workers=normal_workers
            )

            # Process large assets with multipart upload
            for asset, asset_response in large_assets:
                if multipart_executor is None:
                    raise RuntimeError(
                        "Multipart executor not initialized for large assets"
                    )
                futures.append(
                    upload_executor.submit(
                        self._upload_multipart_file_through_signed_url,
                        multipart_executor,
                        upload_session_response.id,
                        asset_response,
                        self.abort_event,
                    )
                )

            # Process small assets with normal upload
            for asset, asset_response in small_assets:
                futures.append(
                    upload_executor.submit(
                        self._upload_file_through_signed_url,
                        asset_response,
                        self.abort_event,
                    )
                )

            success_count = 0
            failed_count = 0
            for future in futures:
                filename, uploaded = future.result()

                if uploaded:
                    logger.debug("Finished Uploading: % s", filename)
                    success_count += 1
                else:
                    logger.debug("Failed to upload %s", filename)
                    failed_count += 1

            logger.debug(
                "Upload session finished: %d success, %d failed",
                success_count,
                failed_count,
            )

        except KeyboardInterrupt as exc:
            self.abort_event.set()

            for future in futures:
                if not future.done():
                    future.cancel()

            self._cancel_upload_session()
            raise KeyboardInterrupt("Upload interrupted by user") from exc

        finally:
            if upload_executor:
                upload_executor.shutdown(wait=False)
            if multipart_executor:
                multipart_executor.shutdown(wait=False)

        return upload_session_response

    def wait_until_done(
        self,
        raise_exception_if: Union[
            OperationStatusOverview, str
        ] = OperationStatusOverview.ERRORED,
    ):
        """
        Wait for all operations to be done.
        This function only works when background is set to False.
        It functions the same as Operation.wait_until_done.

        :param raise_exception_if: The condition to raise error.
        :return: The operation status metadata if the operation has finished,
        """
        assert isinstance(raise_exception_if, (str, OperationStatusOverview))

        if isinstance(raise_exception_if, str):
            raise_exception_if = OperationStatusOverview(raise_exception_if)

        if len(self.operation_ids) == 0:
            logger.debug("All operations finished")
            return True

        with concurrent.futures.ThreadPoolExecutor() as executor:
            # Submit tasks to the executor
            futures = [
                executor.submit(
                    self.operation.wait_until_done,
                    op_id,
                    raise_exception_if=raise_exception_if,
                    abort_event=self.abort_event,
                )
                for op_id in self.operation_ids
            ]

            # Optionally, you can handle the results of the uploads here
            try:
                for future in futures:
                    res = future.result()

                    logger.debug("Finished operation: % s", res)

                logger.debug("All operations finished")

            except KeyboardInterrupt:
                self.abort_event.set()
                for future in futures:
                    if not future.done():
                        future.cancel()
                raise

        return True

    def _check_current_asset_size(self):
        if len(self.assets) >= config.ASSET_UPLOAD_SESSION_BATCH_SIZE:
            upload_operation = self._upload_assets()

            self.operation_ids.append(upload_operation.op_id)
            # clear current batch
            self.assets = []

    def _cancel_upload_session(self) -> None:
        """Cancel upload session"""
        for upload_session_id in self.upload_session_ids:
            self.requester.POST(
                f"/projects/{self.project_id}/assetuploadsessions/{upload_session_id}:cancel",
                response_type=CancelResponse,
            )

    def get_operation_ids(self):
        """
        A list of operation IDs. Because some dependency limits,
        each operation allows a maximum of 1000 assets.
        So if the total number of assets goes up over 1000,
        it will return a list of operation IDs.

        If you want to control the operations manually,
        you can use this function to get the operation ids.
        And the call project.operation.wait_until_done or project.operation.get
        to wait for the operations to finish.

        :return: A list of operation ids.

        :example:
            .. code-block:: python

                ['op_1', 'op_2', 'op_3']

        """
        return self.operation_ids



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   postprocess.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Postprocess utility functions.
"""
# pylint: disable=R0913,R0914,R0915,R0917

import copy
from typing import Any, List, Optional, Tuple

import numpy as np
import pycocotools.mask
from numpy.typing import NDArray


def nms(
    boxes: np.ndarray,
    classes: np.ndarray,
    scores: np.ndarray,
    masks: Optional[np.ndarray] = None,
    keypoints: Optional[np.ndarray] = None,
    keypoint_confidences: Optional[np.ndarray] = None,
    iou_threshold: float = 0.1,
    confidence: float = 0.1,
    sigma: float = 0.5,
) -> Tuple[
    NDArray[Any],
    NDArray[Any],
    NDArray[Any],
    Optional[NDArray[Any]],
    Optional[NDArray[Any]],
    Optional[NDArray[Any]],
]:
    """Carry out non-max supression on the predictions.

    Args:
        boxes: Bounding boxes.
        classes: Classes.
        scores: Scores.
        masks: Masks.
        keypoints: Keypoints.
        keypoint_confidences: Keypoint confidences.
        iou_threshold: IoU threshold.
        confidence: Confidence threshold.
        sigma: Sigma for soft-NMS.

    Returns:
        Filtered boxes, classes, scores, optional masks,
        optional keypoints, and optional keypoint confidences.
    """
    is_soft = False
    use_exp = False

    (
        nms_boxes,
        nms_classes,
        nms_scores,
        nms_masks,
        nms_keypoints,
        nms_keypoint_confidences,
    ) = ([], [], [], [], [], [])

    for each_class in set(classes):
        # handle data for one class
        indices = np.where(classes == each_class)
        filtered_boxes = boxes[indices]
        filtered_classes = classes[indices]
        filtered_scores = scores[indices]
        filtered_masks = masks[indices] if masks is not None else None
        filtered_keypoints = keypoints[indices] if keypoints is not None else None
        filtered_keypoint_confidences = (
            keypoint_confidences[indices] if keypoint_confidences is not None else None
        )

        # make a data copy to avoid breaking
        # during nms operation
        boxes_nms = copy.deepcopy(filtered_boxes)
        classes_nms = copy.deepcopy(filtered_classes)
        scores_nms = copy.deepcopy(filtered_scores)
        masks_nms = (
            copy.deepcopy(filtered_masks) if filtered_masks is not None else None
        )
        keypoints_nms = (
            copy.deepcopy(filtered_keypoints)
            if filtered_keypoints is not None
            else None
        )
        keypoint_confidences_nms = (
            copy.deepcopy(filtered_keypoint_confidences)
            if filtered_keypoint_confidences is not None
            else None
        )

        while len(scores_nms) > 0:
            # pick the max box and store, here
            # we also use copy to persist result
            i = np.argmax(scores_nms, axis=-1)
            nms_boxes.append(copy.deepcopy(boxes_nms[i]))
            nms_classes.append(copy.deepcopy(classes_nms[i]))
            nms_scores.append(copy.deepcopy(scores_nms[i]))

            if masks_nms is not None:
                nms_masks.append(copy.deepcopy(masks_nms[i]))

            if keypoints_nms is not None and keypoint_confidences_nms is not None:
                nms_keypoints.append(copy.deepcopy(keypoints_nms[i]))
                nms_keypoint_confidences.append(
                    copy.deepcopy(keypoint_confidences_nms[i])
                )

            # swap the max line and first line
            boxes_nms[[i, 0], :] = boxes_nms[[0, i], :]
            classes_nms[[i, 0]] = classes_nms[[0, i]]
            scores_nms[[i, 0]] = scores_nms[[0, i]]

            if masks_nms is not None:
                masks_nms[[i, 0]] = masks_nms[[0, i]]

            if keypoints_nms is not None and keypoint_confidences_nms is not None:
                keypoints_nms[[i, 0]] = keypoints_nms[[0, i]]
                keypoint_confidences_nms[[i, 0]] = keypoint_confidences_nms[[0, i]]

            iou = box_diou(boxes_nms)

            # drop the last line since it has been record
            boxes_nms = boxes_nms[1:]
            classes_nms = classes_nms[1:]
            scores_nms = scores_nms[1:]
            masks_nms = masks_nms[1:] if masks_nms is not None else None
            keypoints_nms = keypoints_nms[1:] if keypoints_nms is not None else None
            keypoint_confidences_nms = (
                keypoint_confidences_nms[1:]
                if keypoint_confidences_nms is not None
                else None
            )

            if is_soft:
                # Soft-NMS
                if use_exp:
                    # score refresh formula:
                    # score = score * exp(-(iou^2)/sigma)
                    scores_nms = scores_nms * np.exp(-(iou * iou) / sigma)
                else:
                    # score refresh formula:
                    # score = score * (1 - iou) if iou > threshold
                    depress_mask = np.where(iou > iou_threshold)[0]
                    scores_nms[depress_mask] = scores_nms[depress_mask] * (
                        1 - iou[depress_mask]
                    )
                keep_mask = np.where(scores_nms >= confidence)[0]
            else:
                # normal Hard-NMS
                keep_mask = np.where(iou <= iou_threshold)[0]

            # keep needed box for next loop
            boxes_nms = boxes_nms[keep_mask]
            classes_nms = classes_nms[keep_mask]
            scores_nms = scores_nms[keep_mask]
            masks_nms = masks_nms[keep_mask] if masks_nms is not None else None
            keypoints_nms = (
                keypoints_nms[keep_mask] if keypoints_nms is not None else None
            )
            keypoint_confidences_nms = (
                keypoint_confidences_nms[keep_mask]
                if keypoint_confidences_nms is not None
                else None
            )

    # reformat result for output
    nms_boxes_np = np.array(nms_boxes)
    nms_classes_np = np.array(nms_classes)
    nms_scores_np = np.array(nms_scores)
    nms_masks_np = np.array(nms_masks) if nms_masks else None
    nms_keypoints_np = np.array(nms_keypoints) if nms_keypoints else None
    nms_keypoint_confidences_np = (
        np.array(nms_keypoint_confidences) if nms_keypoint_confidences else None
    )

    return (
        nms_boxes_np,
        nms_classes_np,
        nms_scores_np,
        nms_masks_np,
        nms_keypoints_np,
        nms_keypoint_confidences_np,
    )


def box_diou(boxes):
    """
    Calculate DIoU value of 1st box with other boxes of a box array
    Reference Paper:
        "Distance-IoU Loss: Faster and Better Learning for
        Bounding Box Regression"
        https://arxiv.org/abs/1911.08287

    Args:
      boxes: bbox numpy array, shape=(N, 4), xywh
             x,y are top left coordinates

    Returns:
      diou: numpy array, shape=(N-1,)
            IoU value of boxes[1:] with boxes[0]
    """
    # get box coordinate and area
    x_pos = boxes[:, 0]
    y_pos = boxes[:, 1]
    wid = boxes[:, 2]
    hei = boxes[:, 3]
    areas = wid * hei

    # check IoU
    inter_xmin = np.maximum(x_pos[1:], x_pos[0])
    inter_ymin = np.maximum(y_pos[1:], y_pos[0])
    inter_xmax = np.minimum(x_pos[1:] + wid[1:], x_pos[0] + wid[0])
    inter_ymax = np.minimum(y_pos[1:] + hei[1:], y_pos[0] + hei[0])

    inter_w = np.maximum(0.0, inter_xmax - inter_xmin + 1)
    inter_h = np.maximum(0.0, inter_ymax - inter_ymin + 1)

    inter = inter_w * inter_h
    iou = inter / (areas[1:] + areas[0] - inter)

    # box center distance
    x_center = x_pos + wid / 2
    y_center = y_pos + hei / 2
    center_distance = np.power(x_center[1:] - x_center[0], 2) + np.power(
        y_center[1:] - y_center[0], 2
    )

    # get enclosed area
    enclose_xmin = np.minimum(x_pos[1:], x_pos[0])
    enclose_ymin = np.minimum(y_pos[1:], y_pos[0])
    enclose_xmax = np.maximum(x_pos[1:] + wid[1:], x_pos[0] + wid[0])
    enclose_ymax = np.maximum(y_pos[1:] + wid[1:], y_pos[0] + wid[0])
    enclose_w = np.maximum(0.0, enclose_xmax - enclose_xmin + 1)
    enclose_h = np.maximum(0.0, enclose_ymax - enclose_ymin + 1)
    # get enclosed diagonal distance
    enclose_diagonal = np.power(enclose_w, 2) + np.power(enclose_h, 2)
    # calculate DIoU, add epsilon in denominator to avoid dividing by 0
    diou = iou - 1.0 * (center_distance) / (enclose_diagonal + np.finfo(float).eps)

    return diou


def box_xywh2xyxyn(boxes_xywh: NDArray[Any], height: int, width: int):
    """Convert boxes from xywh to xyxyn format"""
    boxes_xyxyn = np.copy(boxes_xywh)
    boxes_xyxyn[..., 0] = (
        boxes_xywh[..., 0] - boxes_xywh[..., 2] / 2
    ) / width  # top left x
    boxes_xyxyn[..., 1] = (
        boxes_xywh[..., 1] - boxes_xywh[..., 3] / 2
    ) / height  # top left y
    boxes_xyxyn[..., 2] = (
        boxes_xywh[..., 0] + boxes_xywh[..., 2] / 2
    ) / width  # bottom right x
    boxes_xyxyn[..., 3] = (
        boxes_xywh[..., 1] + boxes_xywh[..., 3] / 2
    ) / height  # bottom right y
    boxes_xyxyn = np.clip(boxes_xyxyn, 0, 1)
    return boxes_xyxyn


def clamp(value: float, minimum=0.0, maximum=1.0) -> float:
    """Clamp a value between a minimum and maximum value.

    :param x: Value to clamp.
    :param minimum: Minimum value to clamp to.
    :param maximum: Maximum value to clamp to.
    :return: Clamped value.
    """
    return max(minimum, min(value, maximum))


def get_instance_mask(mask: np.ndarray, threshold: int) -> np.ndarray:
    """Convert class mask to instaance mask"""
    instance_mask = np.zeros_like(mask, np.uint8)
    instance_mask[np.where(mask > threshold)] = 1
    return instance_mask


def get_minimum_bounding_box(mask: np.ndarray) -> List[List[float]]:
    """Get minimum bounding box of a mask and return its 4 normalized vertices.

    Args:
        mask: 2D binary mask array

    Returns:
        List of 4 vertices coordinates in order:
        (v1_top_left, v2_top_right, v3_bottom_right, v4_bottom_left)
    """
    xmin, ymin = np.min(np.where(mask > 0), axis=1)
    xmax, ymax = np.max(np.where(mask > 0), axis=1)

    # Get dimensions for normalization
    height, width = mask.shape

    # Normalize coordinates between 0 and 1
    xmin, xmax = float(xmin) / width, float(xmax) / width
    ymin, ymax = float(ymin) / height, float(ymax) / height

    return [[xmin, ymin], [xmax, ymin], [xmax, ymax], [xmin, ymax]]


def get_minimum_bounding_cube(
    mask: np.ndarray,
) -> List[List[float]]:
    """Get minimum bounding cube of a mask and return its 8 normalized vertices.

    Args:
        mask: 3D binary mask array

    Returns:
        Tuple of 8 vertices coordinates in order:
        (v1_front_top_left, v2_front_top_right, v3_front_bottom_right, v4_front_bottom_left,
         v5_back_top_left, v6_back_top_right, v7_back_bottom_right, v8_back_bottom_left)
    """
    # Get min/max coordinates
    xmin, ymin, zmin = np.min(np.where(mask > 0), axis=1)
    xmax, ymax, zmax = np.max(np.where(mask > 0), axis=1)

    # Get dimensions for normalization
    x_dim, y_dim, z_dim = mask.shape

    # Normalize coordinates between 0 and 1
    xmin, xmax = float(xmin) / x_dim, float(xmax) / x_dim
    ymin, ymax = float(ymin) / y_dim, float(ymax) / y_dim
    zmin, zmax = float(zmin) / z_dim, float(zmax) / z_dim

    # Calculate 8 vertices as lists
    v1 = [xmin, ymin, zmin]  # Front top left
    v2 = [xmax, ymin, zmin]  # Front top right
    v3 = [xmax, ymax, zmin]  # Front bottom right
    v4 = [xmin, ymax, zmin]  # Front bottom left
    v5 = [xmin, ymin, zmax]  # Back top left
    v6 = [xmax, ymin, zmax]  # Back top right
    v7 = [xmax, ymax, zmax]  # Back bottom right
    v8 = [xmin, ymax, zmax]  # Back bottom left

    return [v1, v2, v3, v4, v5, v6, v7, v8]


def numpy_to_rle(binary_mask: np.ndarray) -> Any:
    """
    Convert a numpy binary mask to RLE encoded string.

    Args:
        binary_mask: numpy array of shape (height, width) with values 0 and 1
                    or boolean values

    Returns:
        dict: RLE encoded mask in COCO format
    """
    # Ensure the mask is in the correct format (uint8 with values 0 and 1)
    if binary_mask.dtype == bool:
        binary_mask = binary_mask.astype(np.uint8)
    elif binary_mask.dtype != np.uint8:
        binary_mask = (binary_mask > 0).astype(np.uint8)

    # pycocotools expects Fortran order (column-major)
    binary_mask = np.asfortranarray(binary_mask)

    # Encode to RLE
    rle = pycocotools.mask.encode(binary_mask)

    # Convert counts to string if it's bytes (for JSON serialization)
    if isinstance(rle["counts"], bytes):
        rle["counts"] = rle["counts"].decode("utf-8")

    return rle



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   operation.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Operation API
"""

import logging
import threading
import time
from datetime import datetime, timedelta
from typing import Union

from datature.nexus import config, error, models
from datature.nexus.api.types import OperationStatusOverview
from datature.nexus.client_context import RestContext

logger = logging.getLogger("datature-nexus")


class Operation(RestContext):
    """Datature Operation API Resource."""

    def get(self, op_id: str) -> models.Operation:
        """Retrieves an executed operation status using the operation link.

        :param op_id: The id of the operation as a string.
        :return: A msgspec struct containing the operation metadata with the following structure.

                .. code-block:: python

                        Operation(
                            id='op_d52ab4e6-7760-4d12-9c43-7fc1b8ce1616',
                            kind='nexus.annotations.export',
                            status=OperationStatus(
                                overview='Finished',
                                message='Operation finished',
                                progress=OperationProgress(
                                    unit='whole operation',
                                    with_status=OperationProgressWithStatus(
                                        Queued=0,
                                        Running=0,
                                        Finished=1,
                                        Cancelled=0,
                                        Errored=0
                                    )
                                )
                            ),
                            create_date=1701927649302,
                            update_date=1701927649302
                        )

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")

                        project.operations.get("op_508fc5d1-e908-486d-9e7b-1dca99b80024")
        """
        assert isinstance(op_id, str)
        return self.requester.GET(
            f"/projects/{self.project_id}/operations/{op_id}",
            response_type=models.Operation,
        )

    def wait_until_done(
        self,
        op_id: str,
        timeout: int = config.OPERATION_LOOPING_TIMEOUT_SECONDS,
        raise_exception_if: Union[
            OperationStatusOverview, str
        ] = OperationStatusOverview.ERRORED,
        abort_event: threading.Event = threading.Event(),
    ) -> models.Operation:
        """Continuously retrieves an executed operation status for a
        specified number of times at a specified delay interval.

        Can be used to poll a running operation to monitor execution status.

        :param op_id: The id of the operation as a string.
        :param timeout: The maximum number of times to loop the operation retrieval.
        :param raise_exception_if: The condition to raise error.
        :return: The operation status metadata if the operation has finished,
            a BadRequestError if the operation has errored out,
            or None if the operation is still running.

                .. code-block:: python

                    Operation(
                        id='op_d52ab4e6-7760-4d12-9c43-7fc1b8ce1616',
                        kind='nexus.annotations.export',
                        status=OperationStatus(
                            overview='Finished',
                            message='Operation finished',
                            progress=OperationProgress(
                                unit='whole operation',
                                with_status=OperationProgressWithStatus(
                                    Queued=0,
                                    Running=0,
                                    Finished=1,
                                    Cancelled=0,
                                    Errored=0
                                )
                            )
                        ),
                        create_date=1701927649302,
                        update_date=1701927649302
                    )

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")

                        project.operations.wait_until_done(
                            "op_508fc5d1-e908-486d-9e7b-1dca99b80024"
                        )
        """
        assert isinstance(op_id, str)
        assert isinstance(raise_exception_if, (str, OperationStatusOverview))

        if isinstance(raise_exception_if, str):
            raise_exception_if = OperationStatusOverview(raise_exception_if)

        timeout = timeout or config.OPERATION_LOOPING_TIMEOUT_SECONDS
        response = None
        elapsed_time = datetime.now() + timedelta(seconds=timeout)

        while True:
            response = self.requester.GET(
                f"/projects/{self.project_id}/operations/{op_id}",
                response_type=models.Operation,
            )

            logger.debug("Operation status: %s ", response.status)

            if response.status.overview == OperationStatusOverview.FINISHED.value:
                return response

            if response.status.overview == raise_exception_if.value:
                logger.warning(
                    "Operation %s %s: please contact support at support@datature.io",
                    raise_exception_if.value,
                    op_id,
                )

                raise error.BadRequestError(
                    f"Operation {raise_exception_if.value} {op_id}: "
                    "please contact support at support@datature.io"
                )

            # if the operation has not finished when the timeouts
            if elapsed_time < datetime.now():
                logger.warning(
                    "Operation timeout: please use operation.get(%s) to get the status",
                    op_id,
                )
                return response

            if abort_event.is_set():
                return response

            time.sleep(config.OPERATION_LOOPING_DELAY_SECONDS)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   base_predictor.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Base class for all predictors.
"""

from abc import ABC, abstractmethod
from typing import Any

from datature.nexus.cli.deploy.local.utils.model import Model


class BasePredictor(ABC):
    """Base class for all predictors."""

    _model: Model

    def __init__(self, model: Model, **kwargs):
        """Initialize Base Predictor."""
        self.__dict__.update(kwargs)
        self._model = model

    def run(self, **kwargs) -> dict[str, Any]:
        """Run prediction pipeline."""
        preprocessed_inputs = self._preprocess(**kwargs)
        raw_predictions = self._predict(**preprocessed_inputs, **kwargs)
        postprocessed_predictions = self._postprocess(**raw_predictions, **kwargs)
        return postprocessed_predictions

    @abstractmethod
    def _predict(self, **kwargs) -> Any:
        """Predict on image.

        Args:
            img: Image to predict on.

        Returns:
            Raw model output.
        """
        raise ValueError("Abstract function needs to be implemented")

    @abstractmethod
    def _preprocess(self, **kwargs) -> Any:
        """Preprocess image."""
        raise ValueError("Abstract function needs to be implemented")

    @abstractmethod
    def _postprocess(self, **kwargs) -> dict[str, Any]:
        """Postprocess raw model output into interpretable predictions."""
        raise ValueError("Abstract function needs to be implemented")

    @property
    def model(self) -> Model:
        """Get model."""
        return self._model



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   module.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Module for Saver output.
"""

# pylint: disable=R0911,R0913,R0914,R0917

import base64
import os
import tempfile
from typing import Dict, List, Optional

import cv2
import nibabel as nib
import numpy as np
from scipy.ndimage import zoom

from datature.nexus.cli.deploy.local.types.prediction_request import ResponseFormat
from datature.nexus.cli.deploy.local.types.prediction_response import (
    BitmaskCOCORLE,
    BitmaskNiiGz,
    Prediction,
    Tag,
)
from datature.nexus.cli.deploy.local.types.skeleton import (
    Connection,
    Keypoint,
    Skeleton,
)
from datature.nexus.cli.deploy.local.utils.inference import PredictionTaskType
from datature.nexus.cli.deploy.local.utils.label_map import LabelMap
from datature.nexus.cli.deploy.local.utils.postprocess import (
    get_minimum_bounding_box,
    get_minimum_bounding_cube,
    numpy_to_rle,
)
from datature.nexus.error import Error


class PredictionFormatter:
    """Prediction formatter class."""

    def run(
        self,
        task_type: PredictionTaskType,
        response_format: ResponseFormat,
        image: np.ndarray,
        predictions: Dict[str, np.ndarray],
        label_map: LabelMap,
        skeletons: Optional[Dict[str, Skeleton]] = None,
        affine: Optional[np.ndarray] = None,
    ) -> List[Prediction]:
        """Format predictions in Datature JSON Lines format.

        Args:
            task_type: Task type.
            response_format: Response format.
            image: Image.
            predictions: Dictionary of predictions.
            label_map: Label map object.
            skeletons: Skeletons object.
            affine: Affine matrix.

        Returns:
            Formatted predictions in Datature JSON Lines format
            and optionally NiiGz encoded prediction as base64 string.
        """

        if task_type == PredictionTaskType.IMAGE_CLASSIFICATION:
            return self._format_class_predictions(predictions, label_map)

        if task_type == PredictionTaskType.OBJECT_DETECTION:
            return self._format_boxes(predictions, label_map)

        if task_type == PredictionTaskType.INSTANCE_SEGMENTATION:
            return self._format_instance_masks(image, predictions, label_map)

        if task_type == PredictionTaskType.SEMANTIC_SEGMENTATION:
            if response_format == ResponseFormat.BITMASK:
                return self._format_semantic_masks_to_bitmask(predictions, label_map)

            return self._format_semantic_masks_to_polygons(
                image, predictions, label_map
            )

        if task_type == PredictionTaskType.SEMANTIC_3D_SEGMENTATION:
            return self._format_semantic_3d_masks(image, predictions, label_map, affine)

        if task_type == PredictionTaskType.KEYPOINT_DETECTION:
            return self._format_keypoints(predictions, label_map, skeletons)

        raise Error(f"Unsupported task type: {task_type}")

    def _format_class_predictions(
        self, predictions: Dict[str, np.ndarray], label_map: LabelMap
    ) -> List[Prediction]:
        """Format class predictions in Datature JSON Lines format.

        Args:
            predictions: Dictionary of predictions.
            label_map: Label map object.

        Returns:
            Formatted predictions in Datature JSON Lines format.
        """
        prediction_id = 0
        formatted_predictions = []

        for each_class, each_score in zip(
            predictions["classes"], predictions["scores"]
        ):
            formatted_prediction = Prediction(
                annotation_id=0,
                confidence=float(each_score),
                tag=Tag(
                    id=int(each_class),
                    name=label_map[int(each_class)].name,
                ),
                bound_type="classification",
                contour_type=None,
                bound=None,
            )
            formatted_predictions.append(formatted_prediction)
            prediction_id += 1

        return formatted_predictions

    def _format_boxes(
        self, predictions: Dict[str, np.ndarray], label_map: LabelMap
    ) -> List[Prediction]:
        """Format boxes in Datature JSON Lines format.

        Args:
            predictions: Dictionary of predictions.
            label_map: Label map object.

        Returns:
            Formatted predictions in Datature JSON Lines format.
        """
        prediction_id = 0
        formatted_predictions = []

        for each_box, each_class, each_score in zip(
            predictions["boxes"], predictions["classes"], predictions["scores"]
        ):
            formatted_prediction = Prediction(
                annotation_id=prediction_id,
                confidence=float(each_score),
                tag=Tag(
                    id=int(each_class),
                    name=label_map[int(each_class)].name,
                ),
                bound_type="rectangle",
                contour_type=None,
                bound=[
                    [float(each_box[1]), float(each_box[0])],  # xmin, ymin
                    [float(each_box[1]), float(each_box[2])],  # xmin, ymax
                    [float(each_box[3]), float(each_box[2])],  # xmax, ymax
                    [float(each_box[3]), float(each_box[0])],  # xmax, ymin
                ],
            )
            formatted_predictions.append(formatted_prediction)
            prediction_id += 1

        return formatted_predictions

    def _format_instance_masks(
        self,
        image: np.ndarray,
        predictions: Dict[str, np.ndarray],
        label_map: LabelMap,
    ) -> List[Prediction]:
        """Format instance masks in Datature JSON Lines format.

        Args:
            image: Image.
            predictions: Dictionary of predictions.
            label_map: Label map object.

        Returns:
            Formatted predictions in Datature JSON Lines format.
        """
        prediction_id = 0
        formatted_predictions = []

        for each_box, each_mask, each_class, each_score in zip(
            predictions["boxes"],
            predictions["masks"],
            predictions["classes"],
            predictions["scores"],
        ):
            mask_height, mask_width = each_mask.shape

            xmin, ymin, xmax, ymax = each_box
            ymin = int(ymin * mask_height)
            ymax = int(ymax * mask_height)
            xmin = int(xmin * mask_width)
            xmax = int(xmax * mask_width)

            row = np.expand_dims(np.arange(mask_width), 0)
            col = np.expand_dims(np.arange(mask_height), -1)
            each_mask *= (row >= xmin) * (row <= xmax) * (col >= ymin) * (col <= ymax)

            each_mask = cv2.resize(
                each_mask,
                (image.shape[1], image.shape[0]),
                interpolation=cv2.INTER_LINEAR,
            )
            binary_mask = np.where(each_mask > 0, 255, 0).astype(np.uint8)

            # convert masks to polygon
            contours, _ = cv2.findContours(
                binary_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
            )

            for contour in contours:
                normalized_contour = contour.squeeze().tolist()
                normalized_contour = [
                    [
                        float(normalized_contour[i][0]) / image.shape[1],
                        float(normalized_contour[i][1]) / image.shape[0],
                    ]
                    for i in range(len(normalized_contour))
                ]

                formatted_prediction = Prediction(
                    annotation_id=prediction_id,
                    confidence=float(each_score),
                    tag=Tag(
                        id=int(each_class),
                        name=label_map[int(each_class)].name,
                    ),
                    bound_type="polygon",
                    contour_type="polygon",
                    bound=normalized_contour,
                )
                formatted_predictions.append(formatted_prediction)
                prediction_id += 1

        return formatted_predictions

    def _format_semantic_masks_to_polygons(
        self,
        image: np.ndarray,
        predictions: Dict[str, np.ndarray],
        label_map: LabelMap,
    ) -> List[Prediction]:
        """Format semantic masks in Datature JSON Lines format.

        Args:
            image: Image.
            predictions: Dictionary of predictions.
            label_map: Label map object.

        Returns:
            Formatted predictions in Datature JSON Lines format.
        """
        prediction_id = 0
        formatted_predictions = []

        mask = predictions["masks"]
        scores = predictions["scores"]

        if mask is None:
            return []

        unique_classes = np.unique(mask)
        if 0 in unique_classes:
            unique_classes = unique_classes[unique_classes != 0]

        resized_mask = cv2.resize(
            mask,
            (image.shape[1], image.shape[0]),
            interpolation=cv2.INTER_LINEAR,
        )

        # Process each class
        for class_id in unique_classes:
            if class_id not in label_map.ids.keys():
                continue

            # Create binary mask for this class
            binary_mask = (resized_mask == class_id).astype(np.uint8)

            # Find contours for this class
            contours, _ = cv2.findContours(
                binary_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
            )

            # Get average confidence score for pixels of this class
            class_indices = np.where(mask == class_id)[0]
            if len(class_indices) > 0:
                # Use the scores from the filtered indices that also match this class
                filtered_indices = class_indices
                score = float(np.mean(scores[filtered_indices]))
            else:
                score = 1.0

            # Process each contour
            for contour in contours:
                if len(contour) < 3:
                    continue

                normalized_contour = contour.squeeze().tolist()
                normalized_contour = [
                    [
                        float(normalized_contour[i][0]) / image.shape[1],
                        float(normalized_contour[i][1]) / image.shape[0],
                    ]
                    for i in range(len(normalized_contour))
                ]

                # Format contour into required structure
                formatted_prediction = Prediction(
                    annotation_id=prediction_id,
                    confidence=float(score),
                    tag=Tag(
                        id=int(class_id),
                        name=label_map[int(class_id)].name,
                    ),
                    bound_type="polygon",
                    contour_type="polygon",
                    bound=normalized_contour,
                )

                formatted_predictions.append(formatted_prediction)
                prediction_id += 1
        return formatted_predictions

    def _format_semantic_masks_to_bitmask(
        self,
        predictions: Dict[str, np.ndarray],
        label_map: LabelMap,
    ) -> List[Prediction]:
        """Format semantic masks to COCO-RLE encoded bitmask in Datature JSON Lines format.

        Args:
            predictions: Dictionary of predictions.
            label_map: Label map object.

        Returns:
            Formatted predictions in Datature JSON Lines format.
        """
        prediction_id = 0
        formatted_predictions = []

        mask = predictions["masks"]
        scores = predictions["scores"]

        if mask is None:
            return []

        unique_classes = np.unique(mask)

        if 0 in unique_classes:
            unique_classes = unique_classes[unique_classes != 0]

        # Process each class
        for class_id in unique_classes:
            if class_id not in label_map.ids.keys():
                continue

            # Create binary mask for this class
            binary_mask = (mask == class_id).astype(np.uint8)

            # Get bounding box 2D for this class
            bounding_box = get_minimum_bounding_box(binary_mask)

            rle_string = numpy_to_rle(binary_mask)

            # Get average confidence score for pixels of this class
            class_indices = np.where(mask == class_id)[0]
            if len(class_indices) > 0:
                # Use the scores from the filtered indices that also match this class
                filtered_indices = class_indices
                score = float(np.mean(scores[filtered_indices]))
            else:
                score = 1.0

            formatted_prediction = Prediction(
                annotation_id=prediction_id,
                confidence=float(score),
                tag=Tag(
                    id=int(class_id),
                    name=label_map[int(class_id)].name,
                ),
                bound_type="masks",
                contour_type="bitmask",
                bound=bounding_box,
                bitmask=BitmaskCOCORLE(
                    height=int(mask.shape[0]),
                    width=int(mask.shape[1]),
                    counts=rle_string["counts"],
                ),
            )

            formatted_predictions.append(formatted_prediction)
            prediction_id += 1
        return formatted_predictions

    def _format_semantic_3d_masks(
        self,
        image: np.ndarray,
        predictions: Dict[str, np.ndarray],
        label_map: LabelMap,
        affine: Optional[np.ndarray] = None,
    ) -> List[Prediction]:
        """Format semantic 3D masks in Datature JSON Lines format.

        Args:
            image: Image.
            predictions: Dictionary of predictions.
            label_map: Label map object.
            affine: Affine matrix.

        Returns:
            Formatted predictions in Datature JSON Lines format.
            NiiGz encoded prediction as base64 string.
        """
        prediction_id = 0
        formatted_predictions = []

        mask = predictions["masks"]
        scores = predictions["scores"]

        if mask is None:
            return []

        unique_classes = np.unique(mask)
        if 0 in unique_classes:
            unique_classes = unique_classes[unique_classes != 0]

        zoom_factors = [image.shape[i] / mask.shape[i] for i in range(3)]
        zoomed_mask = np.array(zoom(mask, zoom_factors, order=0))

        for class_id in unique_classes:
            if class_id not in label_map.ids.keys():
                continue

            binary_mask = (zoomed_mask == class_id).astype(np.uint8)

            # get minimum bounding cube
            bounding_cube = get_minimum_bounding_cube(binary_mask)

            # Get all indices where mask equals class_id across all dimensions
            class_indices = np.where(mask == class_id)[0]
            if len(class_indices) > 0:
                # Get scores for all matching voxels
                score = float(np.mean(scores[class_indices]))
            else:
                score = 1.0

            with tempfile.NamedTemporaryFile(
                suffix=".nii.gz", delete=False
            ) as temp_file:
                temp_path = temp_file.name

            try:
                nifti_mask = nib.Nifti1Image(binary_mask, affine)
                nib.save(nifti_mask, temp_path)

                with open(temp_path, "rb") as f:
                    nifti_bytes = f.read()

                base64_encoded_prediction = base64.b64encode(nifti_bytes).decode(
                    "utf-8"
                )

            finally:
                if os.path.exists(temp_path):
                    os.remove(temp_path)

            formatted_prediction = Prediction(
                annotation_id=prediction_id,
                confidence=float(score),
                tag=Tag(id=int(class_id), name=label_map[int(class_id)].name),
                bound_type="masks3d",
                bound=bounding_cube,
                contour_type="bitmask",
                bitmask=BitmaskNiiGz(file=base64_encoded_prediction),
            )
            formatted_predictions.append(formatted_prediction)
            prediction_id += 1

        return formatted_predictions

    def _format_keypoints(
        self,
        predictions: Dict[str, np.ndarray],
        label_map: LabelMap,
        skeletons: Optional[Dict[str, Skeleton]] = None,
    ) -> List[Prediction]:
        """Format keypoints in Datature JSON Lines format.

        Args:
            predictions: Dictionary of predictions.
            label_map: Label map object.
            skeletons: Skeletons object.

        Returns:
            Formatted predictions in Datature JSON Lines format.
        """
        prediction_id = 0
        formatted_predictions = []

        for (
            each_box,
            each_keypoint,
            keypoint_confidences,
            each_class,
            each_score,
        ) in zip(
            predictions["boxes"],
            predictions["keypoints"],
            predictions["keypoint_confidences"],
            predictions["classes"],
            predictions["scores"],
        ):
            class_name = label_map[int(each_class)].name

            keypoints = each_keypoint.reshape(-1, 3)
            keypoints = keypoints[keypoints[:, 2] > 0.5]

            formatted_prediction = Prediction(
                annotation_id=prediction_id,
                confidence=float(each_score),
                tag=Tag(
                    id=int(each_class),
                    name=class_name,
                ),
                bound_type="keypoints",
                contour_type=None,
                bound=[
                    [float(each_box[0]), float(each_box[1])],
                    [float(each_box[2]), float(each_box[1])],
                    [float(each_box[2]), float(each_box[3])],
                    [float(each_box[0]), float(each_box[3])],
                ],
                skeleton=(
                    Skeleton(
                        name=skeletons[class_name].name,
                        keypoints=[
                            Keypoint(name=keypoint.name, category=keypoint.category)
                            for keypoint in skeletons[class_name].keypoints
                        ],
                        connections=[
                            Connection(pair=connection.pair)
                            for connection in skeletons[class_name].connections
                        ],
                    ).to_json()
                    if skeletons
                    else None
                ),
                keypoint_confidences=keypoint_confidences.tolist(),
                keypoints=keypoints.tolist(),
            )
            formatted_predictions.append(formatted_prediction)
            prediction_id += 1

        return formatted_predictions



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   manager.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Outpost installation manager module.
"""

# pylint: disable=W0718,E1120

import queue
import sys
import time
import traceback

from InquirerPy import inquirer

from datature.nexus.cli.outpost import consts as CONSTS
from datature.nexus.cli.outpost import utils
from datature.nexus.cli.outpost.types import (
    DeviceStatusMessage,
    OutpostDeviceConfig,
    RegistrationStatus,
)
from datature.nexus.error import Error

ACTION_TIMEOUT_SECONDS = 300


class OutpostManager:
    """Outpost Manager class."""

    def __init__(self):
        CONSTS.OUTPOST_CONFIG_ROOT_DIR.mkdir(parents=True, exist_ok=True)
        self._outpost_device_config = OutpostDeviceConfig.read_from_file()

    def create_configuration(self):
        """Create Outpost configuration."""
        try:
            utils.create_configuration(self._outpost_device_config)
            self._outpost_device_config.write_to_file()
        except KeyboardInterrupt:
            sys.exit(1)
        except Error:
            sys.exit(1)
        except Exception:
            traceback.print_exc()
            sys.exit(1)

    def install(self):
        """Install Outpost on device."""
        try:
            utils.prompt_for_configuration_version(self._outpost_device_config)
            utils.prompt_for_runtime_config(self._outpost_device_config)
            print()

            utils.register_device(self._outpost_device_config)
            utils.setup_outpost_credentials(self._outpost_device_config)
            utils.get_outpost_device_config_files(self._outpost_device_config)
            utils.get_outpost_runtime_files(self._outpost_device_config)

            self._outpost_device_config.write_to_file()

            print("Installing Outpost runtime... This may take a few minutes.")
            utils.install_runtime(self._outpost_device_config)

        except KeyboardInterrupt:
            sys.exit(1)

        except Error as exc:
            print(exc)
            sys.exit(1)

        except Exception:
            traceback.print_exc()
            sys.exit(1)

    def uninstall(self):
        """Uninstall Outpost from device."""
        try:
            self.pause()
            utils.deregister_device(self._outpost_device_config)

        except Error:
            print(
                "Warning: error occurred while deregistering device. "
                "It may have already been deregistered."
            )

        try:
            cleanup = inquirer.confirm(
                message=(
                    "Do you also want to remove all Outpost runtime files from your device?"
                ),
            ).execute()

            if cleanup:
                utils.uninstall_runtime(self._outpost_device_config)

            self._outpost_device_config.remove_from_file()

        except Error as exc:
            print(exc)
            sys.exit(1)

        except Exception:
            traceback.print_exc()
            sys.exit(1)

    @utils.start_spinner(raise_on_error=True)
    def pause(self, message_queue: queue.Queue):
        """Pause Outpost runtime."""
        message_queue.put("Pausing Outpost runtime...")

        runtime_status = utils.get_runtime_status_message(self._outpost_device_config)

        if runtime_status in [DeviceStatusMessage.OFFLINE.value]:
            return (
                "Outpost runtime is offline. "
                "It may have already been paused, or the device is unreachable."
            )

        if runtime_status in [
            DeviceStatusMessage.PAUSED.value,
            DeviceStatusMessage.PAUSED_ROLLING_OUT.value,
        ]:
            return "Outpost runtime is already paused."

        status_data = {"spec": {"registrationStatus": RegistrationStatus.PAUSED.value}}

        utils.patch_device(self._outpost_device_config, status_data)

        start_time = time.time()
        while utils.get_runtime_status_message(self._outpost_device_config) not in [
            DeviceStatusMessage.PAUSED_ROLLING_OUT.value,
            DeviceStatusMessage.PAUSED.value,
        ]:
            if time.time() - start_time > ACTION_TIMEOUT_SECONDS:
                raise TimeoutError("Pausing Outpost runtime timed out.")

            time.sleep(1)

        if (
            utils.get_runtime_status_message(self._outpost_device_config)
            == DeviceStatusMessage.PAUSED_ROLLING_OUT.value
        ):
            return (
                "Outpost runtime initiating pause. "
                "Run `datature outpost status` to check the status."
            )

        return "Outpost runtime paused successfully."

    @utils.start_spinner
    def resume(self, message_queue: queue.Queue):
        """Resume Outpost runtime."""
        message_queue.put("Resuming Outpost runtime...")

        runtime_status = utils.get_runtime_status_message(self._outpost_device_config)

        if runtime_status in [
            DeviceStatusMessage.READY.value,
            DeviceStatusMessage.READY_ROLLING_OUT.value,
        ]:
            return "Outpost runtime is already resumed."

        status_data = {
            "spec": {"registrationStatus": RegistrationStatus.REGISTERED.value}
        }

        utils.patch_device(self._outpost_device_config, status_data)

        start_time = time.time()
        while utils.get_runtime_status_message(self._outpost_device_config) not in [
            DeviceStatusMessage.READY_ROLLING_OUT.value,
            DeviceStatusMessage.READY.value,
        ]:
            if time.time() - start_time > ACTION_TIMEOUT_SECONDS:
                raise TimeoutError("Resuming Outpost runtime timed out.")

            time.sleep(1)

        if (
            utils.get_runtime_status_message(self._outpost_device_config)
            == DeviceStatusMessage.READY_ROLLING_OUT.value
        ):
            return (
                "Outpost runtime initiating resume. "
                "Run `datature outpost status` to check the status."
            )

        return "Outpost runtime resumed successfully."

    def status(self, user: str = "datature") -> None:
        """Get Outpost runtime status."""
        utils.validate_certificate(self._outpost_device_config)
        utils.get_outpost_runtime_status(self._outpost_device_config, user)
        current_device_info = utils.get_current_device_info()

        utils.format_outpost_runtime_status(
            self._outpost_device_config, current_device_info
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   image_decoder.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Image decoder.
"""

import base64
import os
import tempfile
from typing import Any, Dict, Optional

import cv2
import nibabel as nib
import numpy as np

from datature.nexus.cli.deploy.local.types.prediction_request import ResponseFormat


def decode_image_from_base64(
    data: str, response_format: Optional[ResponseFormat] = None
) -> Dict[str, Any]:
    """Decode image from base64 string.

    Args:
        data: The base64 encoded string.
        response_format: The response format.

    Returns:
        Dictionary containing the image and optionally the affine matrix for NIFTI files.
    """
    parts = data.split(",")
    if len(parts) > 1:
        image = parts[1]
    else:
        image = data

    decoded_image = base64.b64decode(image)

    if response_format == ResponseFormat.NIFTI:
        with tempfile.NamedTemporaryFile(suffix=".nii.gz", delete=False) as temp_file:
            temp_file.write(decoded_image)
            temp_path = temp_file.name

        try:
            nii_image = nib.nifti1.load(temp_path)
            return {
                "image": nii_image.get_fdata(),
                "affine": nii_image.affine,
            }

        finally:
            if os.path.exists(temp_path):
                os.remove(temp_path)

    bgr_image = cv2.imdecode(np.frombuffer(decoded_image, np.uint8), cv2.IMREAD_COLOR)
    rgb_image = cv2.cvtColor(bgr_image, cv2.COLOR_BGR2RGB)

    return {"image": rgb_image}


def decode_image_from_path(data: str) -> Dict[str, Any]:
    """Decode image from path.

    Args:
        data: The path to the image.

    Returns:
        Dictionary containing the image and optionally the affine matrix for NIFTI files.
    """
    if data.endswith(".nii.gz"):
        nii_image = nib.nifti1.load(data)

        return {
            "image": nii_image.get_fdata(),
            "affine": nii_image.affine,
        }

    bgr_image = cv2.imread(data)
    rgb_image = cv2.cvtColor(bgr_image, cv2.COLOR_BGR2RGB)

    return {"image": rgb_image}


def decode_image_from_bytes(
    data: bytes, response_format: Optional[ResponseFormat] = None
) -> Dict[str, Any]:
    """Decode image from bytes.

    Args:
        data: The bytes of the image.
        response_format: The response format.

    Returns:
        Dictionary containing the image and optionally the affine matrix for NIFTI files.
    """
    if response_format == ResponseFormat.NIFTI:
        with tempfile.NamedTemporaryFile(suffix=".nii.gz", delete=False) as temp_file:
            temp_file.write(data)
            temp_path = temp_file.name

        try:
            nii_image = nib.nifti1.load(temp_path)
            return {
                "image": nii_image.get_fdata(),
                "affine": nii_image.affine,
            }

        finally:
            if os.path.exists(temp_path):
                os.remove(temp_path)

    bgr_image = cv2.imdecode(np.frombuffer(data, np.uint8), cv2.IMREAD_COLOR)
    rgb_image = cv2.cvtColor(bgr_image, cv2.COLOR_BGR2RGB)

    return {"image": rgb_image}



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   consts.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Constants for local deployment
"""

from pathlib import Path

# Server Configuration
DEPLOY_CONFIG_ROOT_DIR = Path.home() / ".config/datature"
DEPLOY_CONFIGS_FILE_PATH = DEPLOY_CONFIG_ROOT_DIR / "datature_local_deploy_configs"
DEFAULT_PORT = 9449

DEFAULT_ALLOWED_ORIGINS = ["https://nexus.datature.io"]
DEFAULT_ALLOWED_METHODS = ["GET", "POST", "OPTIONS"]
DEFAULT_MAX_AGE = 3600

DEFAULT_RUNTIME = "CPU"
DEFAULT_DEVICES = "auto"

# Model Configuration
DEFAULT_MODEL_DIR = Path.home() / ".datature" / "models"
DEFAULT_TIMEOUT = 120

# Runtime Configuration
RUNTIME_OPTIONS = {
    "CPU": "CPUExecutionProvider",
    "GPU": "CUDAExecutionProvider",
}



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   multipart.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Asset Multipart Handler
"""

import os
from typing import List

from datature.nexus.api.types import AssetFilePart


class MultipartHandler:
    """Handles splitting files for multipart upload."""

    def __init__(self, file_path: str, part_count: int = 1) -> None:
        """
        Initialize the MultipartHandler.

        :param file_path: Path to the file to upload
        :param part_count: Number of parts to split the file into
        """
        self.file_path = file_path
        self.file_size = os.path.getsize(file_path)
        self.part_count = part_count
        self.parts = self._split_file_into_parts()

    def _split_file_into_parts(self) -> List[AssetFilePart]:
        """Split the file into parts for multipart upload."""
        parts = []
        part_size = self.file_size // self.part_count

        for i in range(self.part_count):
            start_byte = i * part_size
            end_byte = (
                start_byte + part_size if i < self.part_count - 1 else self.file_size
            )

            parts.append(
                AssetFilePart(
                    part_number=i + 1,
                    start_byte=start_byte,
                    end_byte=end_byte,
                )
            )

        return parts

    def read_part_data(self, part: AssetFilePart) -> bytes:
        """
        Read the data for a specific part.

        :param part: The part to read
        :return: Bytes data of the part
        """
        with open(self.file_path, "rb") as f:
            f.seek(part.start_byte)
            return f.read(part.end_byte - part.start_byte)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   consts.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Outpost consts.
"""

import os
from pathlib import Path

NEXUS_ENDPOINT = os.getenv("DATATURE_API_BASE_URL", "https://api.datature.io")
MQTT_ENDPOINT = os.getenv(
    "DATATURE_OUTPOST_MQTT_ENDPOINT", "mqtts://mqtt.outpost.datature.io:8883"
)
DOCS_ENV_SETUP_URL = "https://developers.datature.io/docs/outpost-setup"
DOCS_ERROR_HANDLING_URL = "https://developers.datature.io/docs/outpost-error-handling"
SUPPORT_EMAIL = "support@datature.io"

OUTPOST_CONFIG_ROOT_DIR = Path.home() / ".config/datature"
OUTPOST_CONFIG_FILE_PATH = OUTPOST_CONFIG_ROOT_DIR / "datature_outpost_device_config"
OUTPOST_ERROR_LOG_DIR = Path.home() / ".logs"
OUTPOST_ROOT_DIR = Path.home() / ".datature/outpost"

MINIMUM_RAM_BYTES = 4 * 1024**3
MINIMUM_STORAGE_BYTES = 2 * 1024**3

RESERVED_DEVICE_TAGS = ["group", "timezone"]



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██
@File    :   server.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Local deployment server.
"""

# pylint: disable=R0913,R0917

from typing import List, Optional, Union

import litserve as ls
from starlette.middleware.cors import CORSMiddleware

from datature.nexus.cli.deploy.local.api import InferenceAPI
from datature.nexus.cli.deploy.local.consts import (
    DEFAULT_ALLOWED_METHODS,
    DEFAULT_ALLOWED_ORIGINS,
    DEFAULT_MAX_AGE,
    DEFAULT_PORT,
    DEFAULT_RUNTIME,
    DEFAULT_TIMEOUT,
)
from datature.nexus.cli.deploy.local.endpoints import add_custom_endpoints
from datature.nexus.cli.deploy.local.logger import LOGGING_CONFIG


def start_server(
    deployment_id: str,
    secret_key: str,
    project_id: str,
    project_name: str,
    artifact_id: str,
    artifact_name: str,
    runtime: str = DEFAULT_RUNTIME,
    devices: Union[str, int] = "auto",
    port: int = DEFAULT_PORT,
    allowed_origins: Optional[List[str]] = None,
):
    """Start the local deployment server

    Args:
        deployment_id: The deployment ID
        secret_key: The project secret key
        project_id: The project ID
        project_name: The project name
        artifact_id: The artifact ID
        runtime: The runtime to use (CPU or GPU), defaults to CPU
        port: The port to use, defaults to 9449
        allowed_origins: The origins to allow as a comma-separated string,
            defaults to https://nexus.datature.io
    """
    if allowed_origins is None:
        allowed_origins = DEFAULT_ALLOWED_ORIGINS.copy()

    # Initialize the inference API
    api = InferenceAPI(secret_key, project_id, artifact_id, runtime)

    cors_middleware = (
        CORSMiddleware,
        {
            "allow_origins": allowed_origins,
            "allow_methods": DEFAULT_ALLOWED_METHODS,
            "allow_headers": ["*"],
            "allow_credentials": True,
            "max_age": DEFAULT_MAX_AGE,
        },
    )

    server = ls.LitServer(
        api,
        accelerator="cpu" if runtime == DEFAULT_RUNTIME else "cuda",
        devices=devices,
        middlewares=[cors_middleware],
        model_metadata={
            "deployment_id": deployment_id,
            "project_id": project_id,
            "artifact_id": artifact_id,
            "artifact_name": artifact_name,
        },
        track_requests=True,
        timeout=DEFAULT_TIMEOUT,
        healthcheck_path="/datature-local-deployment-ping",
    )

    server = add_custom_endpoints(
        server, project_id.removeprefix("proj_"), project_name
    )

    server.run(
        port=port,
        log_config=LOGGING_CONFIG,
        generate_client_file=False,
        num_api_servers=int(devices) if devices != "auto" else None,
    )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   common.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Common utilities for the local deployment server.
"""

# pylint: disable=W0718

import queue
import sys
import threading
from functools import wraps
from typing import Callable, Optional, Union

from colorama import Fore
from halo import Halo

from datature.nexus.error import Error


def start_spinner(func: Optional[Callable] = None) -> Union[Callable, Callable]:
    """Start spinner for the function."""

    def decorator(func: Callable):
        @wraps(func)
        def wrapper(*args, **kwargs):
            wait_spinner = Halo(spinner="dots")
            message_queue = queue.Queue()
            stop_thread = threading.Event()

            def update_spinner():
                while not stop_thread.is_set():
                    try:
                        new_message = message_queue.get(timeout=0.1)
                        if not kwargs.get("quiet", False) and new_message is not None:
                            wait_spinner.text = new_message
                    except queue.Empty:
                        continue

            thread = threading.Thread(target=update_spinner)
            wait_spinner.start()
            thread.start()

            try:
                result, message = func(*args, **kwargs, message_queue=message_queue)
                if not kwargs.get("quiet", False):
                    wait_spinner.succeed(message)
                return result
            except KeyboardInterrupt:
                wait_spinner.stop()
                sys.exit(0)
            except Error as exc:
                error_symbol = f"{Fore.CYAN}●{Fore.RESET}"
                wait_spinner.stop_and_persist(error_symbol, exc.message)
                sys.exit(1)
            except Exception as exc:
                error_symbol = f"{Fore.CYAN}●{Fore.RESET}"
                wait_spinner.stop_and_persist(error_symbol, str(exc))
                sys.exit(1)
            finally:
                stop_thread.set()
                thread.join()
                wait_spinner.stop()

        return wrapper

    if func is None:
        return decorator
    return decorator(func)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   entropy.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Entropy calculation utility functions.
"""

from typing import Dict, List

import numpy as np

from datature.nexus.cli.deploy.local.types.prediction_response import Prediction
from datature.nexus.cli.deploy.local.utils.label_map import LabelMap


def calculate_avg_entropy(predictions: List[Prediction]) -> float:
    """Calculate average entropy across all prediction scores.

    Args:
        predictions: Dictionary containing prediction data with 'scores' key.
        threshold: Minimum score threshold for entropy calculation.

    Returns:
        Average entropy value.
    """
    scores = [prediction.confidence for prediction in predictions]
    if len(scores) == 0:
        return 0.0

    valid_scores = [score for score in scores if 0 < score <= 1]

    if len(valid_scores) == 0:
        return 0.0

    entropies = [
        -score * np.log2(score) - (1 - score) * np.log2(1 - score)
        for score in valid_scores
    ]

    return float(np.mean(entropies))


def calculate_avg_entropy_per_class(
    predictions: List[Prediction], label_map: LabelMap
) -> Dict[str, float]:
    """Calculate average entropy per class based on prediction scores.

    Args:
        predictions: Dictionary containing prediction data with 'scores' and 'classes' keys.
        label_map: Label map object.
        threshold: Minimum score threshold for entropy calculation.

    Returns:
        Dictionary containing average entropy per class.
    """
    scores = np.array([prediction.confidence for prediction in predictions])
    classes = np.array([prediction.tag.id for prediction in predictions])

    if len(scores) == 0 or len(classes) == 0:
        return {}

    unique_classes = np.unique(classes)
    unique_class_names = [label_map[int(class_id)].name for class_id in unique_classes]

    class_entropies = []

    for class_id in unique_classes:
        # Get scores for this class
        class_mask = classes == class_id
        class_scores = scores[class_mask]

        if len(class_scores) == 0:
            continue

        valid_scores = [score for score in class_scores if 0 < score <= 1]

        if len(valid_scores) == 0:
            continue

        entropies = [
            -score * np.log2(score) - (1 - score) * np.log2(1 - score)
            for score in valid_scores
        ]

        class_entropies.append(np.mean(entropies))

    if len(class_entropies) == 0:
        return {}

    return {
        class_name: float(entropy)
        for class_name, entropy in zip(unique_class_names, class_entropies)
    }



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   object.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Object API
"""

from typing import Union

from datature.nexus import models
from datature.nexus.api.types import ObjectFilter, Pagination
from datature.nexus.client_context import RestContext


class Object(RestContext):
    """Datature Object API Resource."""

    def list(
        self,
        pagination: Union[Pagination, dict, None] = None,
        filters: Union[ObjectFilter, dict, None] = None,
        include_attributes: bool = False,
    ) -> models.PaginationResponse[models.Object]:
        """Lists objects in the project.

        :param pagination: Pagination object.
        :param filters: ObjectFilter object.
        :param include_attributes: Include attributes in the response.

        :return: PaginationResponse of an Object containing a list of
            assets and a list of annotations with the following structure:

                .. code-block:: python

                    PaginationResponse(
                        next_page='NjVhYTE0ZDY3YzQ5MThlMTJjYTUyOGRk',
                        prev_page=None,
                        data=Object(
                            annotations=[
                                [
                                    Annotation(
                                        id='annot_e5028bdc-122e-4837-b354-d82b95320b69',
                                        project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                                        asset_id='asset_6647b58e-86af-460c-9a85-efead830aee8',
                                        tag='dog',
                                        bound_type='Polygon',
                                        bound=[
                                            [0.42918, 0.45322666666666667],
                                            [0.43677999999999995, 0.4579466666666666],
                                            [0.43677999999999995, 0.46941333333333335],
                                            [0.44236000000000003, 0.4788800000000001],
                                            [0.45146, 0.4829333333333333],
                                            [0.45754, 0.48696000000000006],
                                            [0.48234, 0.4842666666666666],
                                            [0.4869, 0.43432000000000004],
                                            [0.50514, 0.4302666666666667]
                                        ],
                                        create_date=None,
                                        ontology_id='ontology_843e486c-58d7-45a7-b722-f4948e204a56',
                                        attributes=[
                                            OntologyAttributeValue(
                                                id='attri_d0877827-63d6-4794-b520-ef3e0c57ef71',
                                                name='Tag Group',
                                                value=['person']
                                            )
                                        ]
                                    ],
                                ],
                                assets=[
                                    Asset(
                                        id='asset_8208740a-2d9c-46e8-abb9-5777371bdcd3',
                                        filename='boat180.png',
                                        project='proj_cd067221d5a6e4007ccbb4afb5966535',
                                        status='None',
                                        create_date=1701927649302,
                                        url='',
                                        metadata=AssetMetadata(
                                            file_size=186497,
                                            mime_type='image/png',
                                            height=243,
                                            width=400,
                                            groups=['main'],
                                            custom_metadata={'captureAt': '2021-03-10T09:00:00Z'}
                                        ),
                                        statistic=AssetAnnotationsStatistic(
                                            tags_count=[],
                                            total_annotations=0
                                    )
                                    )
                                ],
                            ],
                        )
                    )

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")

                        project.objects.list({
                            "limit": 2,
                                "page": "ZjYzYmJkM2FjN2UxOTA4ZmU0ZjE0Yjk5Mg"
                            }
                        )

        """
        assert isinstance(pagination, (Pagination, dict, type(None)))
        assert isinstance(filters, (ObjectFilter, dict, type(None)))

        if isinstance(pagination, dict):
            pagination = Pagination(**pagination)
        if pagination is None:
            pagination = Pagination()

        if isinstance(filters, dict):
            filters = ObjectFilter(**filters)
        if filters is None:
            filters = ObjectFilter()

        return self.requester.GET(
            f"/projects/{self.project_id}/objects",
            query={
                **pagination.to_json(),
                **filters.to_json(),
                "include_attributes": include_attributes,
            },
            response_type=models.PaginationResponse[models.Object],
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   utils.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Outpost utility functions.
"""

# pylint: disable=W0718

import json
import os
import queue
import shutil
import stat
import subprocess
import sys
import tempfile
import zipfile
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, Optional, Tuple
from uuid import uuid4

import msgspec
import requests
import yaml
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from InquirerPy import inquirer

from datature.nexus.cli.outpost import consts as CONSTS
from datature.nexus.cli.outpost import messages as MESSAGES
from datature.nexus.cli.outpost.types import (
    DeviceStatusMessage,
    DeviceStatusProperties,
    MetadataSpec,
    OutpostDeviceConfig,
    OutpostDeviceConfiguration,
    RegistrationStatus,
    RuntimeStatus,
    UpdaterStatus,
)
from datature.nexus.cli.outpost.utils import (
    DEVICE_CONFIG,
    interpolate_config_variables,
    start_spinner,
)
from datature.nexus.config import REQUEST_TIME_OUT_SECONDS
from datature.nexus.error import Error

RUNTIME_CREDENTIALS_DIR = Path("/home/datature/.datature/ssl")
DEFAULT_CREDENTIALS_DIR = Path.home() / ".datature" / "ssl"


def validate_device(
    outpost_device_config: OutpostDeviceConfig, display_name: str
) -> Dict[str, Any]:
    """Validate whether the device name has been taken or not."""
    response = requests.post(
        (
            f"{CONSTS.NEXUS_ENDPOINT}/workspaces/ws_"
            f"{outpost_device_config.device_credentials.workspace_id}"
            f"/outpost/deviceValidation"
        ),
        headers={
            "Secret-Key": outpost_device_config.device_credentials.secret_key,
            "Content-Type": "application/json",
        },
        data=json.dumps({"id": outpost_device_config.id, "displayName": display_name}),
        timeout=REQUEST_TIME_OUT_SECONDS,
    )

    return response.json()


def list_devices(outpost_device_config: OutpostDeviceConfig) -> Dict[str, Any]:
    """List all devices in the workspace."""
    try:
        response = requests.get(
            (
                f"{CONSTS.NEXUS_ENDPOINT}/workspaces/ws_"
                f"{outpost_device_config.device_credentials.workspace_id}"
                f"/outpost/devices?limit=1000"
            ),
            headers={
                "Secret-Key": outpost_device_config.device_credentials.secret_key,
                "Content-Type": "application/json",
            },
            timeout=REQUEST_TIME_OUT_SECONDS,
        )
        response.raise_for_status()

    except requests.exceptions.ConnectionError as exc:
        raise Error(MESSAGES.SERVER_CONNECTION_ERROR_MESSAGE) from exc

    except requests.exceptions.HTTPError as exc:
        if response.status_code == 400:
            raise Error(MESSAGES.INSUFFICIENT_DEVICE_QUOTA_ERROR_MESSAGE) from exc

        if response.status_code == 402:
            raise Error(MESSAGES.FEATURE_NOT_AVAILABLE_ERROR_MESSAGE) from exc

        if response.status_code == 403:
            raise Error(
                MESSAGES.SERVER_FORBIDDEN_ERROR_MESSAGE.format(
                    outpost_device_config.device_credentials.workspace_id
                )
            ) from exc

        raise Error(
            MESSAGES.SERVER_UNKNOWN_ERROR_MESSAGE.format(response.text)
        ) from exc

    return response.json()


def get_device(outpost_device_config: OutpostDeviceConfig) -> Dict[str, Any]:
    """Get the current device registration information."""
    try:
        response = requests.get(
            (
                f"{CONSTS.NEXUS_ENDPOINT}/workspaces/ws_"
                f"{outpost_device_config.device_credentials.workspace_id}"
                f"/outpost/devices/{outpost_device_config.id}"
            ),
            headers={
                "Secret-Key": outpost_device_config.device_credentials.secret_key,
                "Content-Type": "application/json",
            },
            timeout=REQUEST_TIME_OUT_SECONDS,
        )
        response.raise_for_status()

    except requests.exceptions.ConnectionError as exc:
        raise Error(MESSAGES.SERVER_CONNECTION_ERROR_MESSAGE) from exc

    except requests.exceptions.HTTPError as exc:
        if response.status_code == 400:
            raise Error(MESSAGES.INSUFFICIENT_DEVICE_QUOTA_ERROR_MESSAGE) from exc

        if response.status_code == 402:
            raise Error(MESSAGES.FEATURE_NOT_AVAILABLE_ERROR_MESSAGE) from exc

        if response.status_code == 403:
            raise Error(
                MESSAGES.SERVER_FORBIDDEN_ERROR_MESSAGE.format(
                    outpost_device_config.device_credentials.workspace_id
                )
            ) from exc

        raise Error(
            MESSAGES.SERVER_UNKNOWN_ERROR_MESSAGE.format(response.text)
        ) from exc

    return response.json()


def patch_device(
    outpost_device_config: OutpostDeviceConfig, data: Dict[str, Any]
) -> None:
    """Patch the device registration information."""
    if "id" in data:
        del data["id"]

    try:
        response = requests.patch(
            (
                f"{CONSTS.NEXUS_ENDPOINT}/workspaces/ws_"
                f"{outpost_device_config.device_credentials.workspace_id}"
                f"/outpost/devices/{outpost_device_config.id}"
            ),
            headers={
                "Secret-Key": outpost_device_config.device_credentials.secret_key,
                "Content-Type": "application/json",
            },
            data=json.dumps(data),
            timeout=REQUEST_TIME_OUT_SECONDS,
        )
        response.raise_for_status()

    except requests.exceptions.ConnectionError as exc:
        raise Error(MESSAGES.SERVER_CONNECTION_ERROR_MESSAGE) from exc

    except requests.exceptions.HTTPError as exc:
        if response.status_code == 400:
            raise Error(MESSAGES.INSUFFICIENT_DEVICE_QUOTA_ERROR_MESSAGE) from exc

        if response.status_code == 402:
            raise Error(MESSAGES.FEATURE_NOT_AVAILABLE_ERROR_MESSAGE) from exc

        if response.status_code == 403:
            raise Error(
                MESSAGES.SERVER_FORBIDDEN_ERROR_MESSAGE.format(
                    outpost_device_config.device_credentials.workspace_id
                )
            ) from exc

        raise Error(
            MESSAGES.SERVER_UNKNOWN_ERROR_MESSAGE.format(response.text)
        ) from exc


def get_runtime_status_message(outpost_device_config: OutpostDeviceConfig) -> str:
    """Get the status message of the Outpost runtime."""
    return get_device(outpost_device_config).get("status", {}).get("message")


def register_device(outpost_device_config: OutpostDeviceConfig) -> None:
    """Register the current device with the server."""
    registration_data = {
        "id": outpost_device_config.id,
        "name": outpost_device_config.name,
        "spec": {
            "registrationStatus": RegistrationStatus.REGISTERED.value,
            "tags": dict(outpost_device_config.tags),
            "deviceInfo": msgspec.to_builtins(outpost_device_config.device_info),
            "runtime": {
                "kind": "Outpost",
                "version": outpost_device_config.runtime.id,
                "configuration": outpost_device_config.configuration.id,
            },
            "revisionId": str(outpost_device_config.runtime.version),
            "collectedMetrics": [],
        },
    }

    try:
        response = requests.post(
            (
                f"{CONSTS.NEXUS_ENDPOINT}/workspaces/ws_"
                f"{outpost_device_config.device_credentials.workspace_id}/outpost/devices"
            ),
            headers={
                "Secret-Key": outpost_device_config.device_credentials.secret_key,
                "Content-Type": "application/json",
            },
            data=json.dumps(registration_data),
            timeout=REQUEST_TIME_OUT_SECONDS,
        )
        response.raise_for_status()

    except requests.exceptions.ConnectionError as exc:
        raise Error(MESSAGES.SERVER_CONNECTION_ERROR_MESSAGE) from exc

    except requests.exceptions.HTTPError as exc:
        if response.status_code == 400:
            raise Error(MESSAGES.INSUFFICIENT_DEVICE_QUOTA_ERROR_MESSAGE) from exc

        if response.status_code == 402:
            raise Error(MESSAGES.FEATURE_NOT_AVAILABLE_ERROR_MESSAGE) from exc

        if response.status_code == 403:
            raise Error(
                MESSAGES.SERVER_FORBIDDEN_ERROR_MESSAGE.format(
                    outpost_device_config.device_credentials.workspace_id
                )
            ) from exc

        if response.status_code == 409:
            inquirer.confirm(
                message=(
                    f"A device with ID {outpost_device_config.id} "
                    "has already been registered. Do you want to overwrite it?"
                ),
                default=False,
            ).execute()
            patch_device(outpost_device_config, registration_data)
            return

        raise Error(
            MESSAGES.SERVER_UNKNOWN_ERROR_MESSAGE.format(response.text)
        ) from exc

    device_response = response.json()

    device_status = device_response["status"]

    if device_status["message"] != DeviceStatusMessage.REGISTERING.value:
        raise Error("Error: Device registration failed. Please try again.")

    outpost_device_config.uid = device_response["uid"]
    outpost_device_config.metadata_generation = device_response["metadataGeneration"]
    outpost_device_config.write_to_file()


def generate_pem_keys() -> Tuple[bytes, bytes]:
    """Generate a 3072-bit RSA key pair and save them to PEM files."""
    # Generate a 3072-bit RSA key
    private_key = rsa.generate_private_key(
        public_exponent=65537,  # Standard value for RSA
        key_size=3072,
    )

    # Serialize the private key to PEM format
    pem_private_key = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )

    Path(DEFAULT_CREDENTIALS_DIR).mkdir(parents=True, exist_ok=True)

    private_key_path = DEFAULT_CREDENTIALS_DIR / "key.pvt.pem"
    with open(private_key_path, "wb") as f:
        f.write(pem_private_key)

    # Extract the public key
    public_key = private_key.public_key()

    # Serialize the public key to PEM format
    pem_public_key = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )

    public_key_path = DEFAULT_CREDENTIALS_DIR / "key.pub.pem"
    with open(public_key_path, "wb") as f:
        f.write(pem_public_key)

    return pem_private_key, pem_public_key


@start_spinner
def setup_outpost_credentials(
    outpost_device_config: OutpostDeviceConfig, message_queue: queue.Queue
):
    """Make a certificate signing request to the server."""
    message_queue.put("Generating device certificate...")

    certificate_id = f"outpostdevicecertificate_{str(uuid4()).replace('-', '')[-11:]}"
    _, pem_public_key = generate_pem_keys()

    certificate_data = {
        "id": certificate_id,
        "spec": {
            "certificateSigningRequest": {
                "kind": "PublicKey",
                "key": pem_public_key.decode("utf-8").strip(),
            }
        },
    }
    try:
        response = requests.post(
            (
                f"{CONSTS.NEXUS_ENDPOINT}/workspaces/ws_"
                f"{outpost_device_config.device_credentials.workspace_id}"
                f"/outpost/devices/{outpost_device_config.id}/certificates"
            ),
            headers={
                "Secret-Key": outpost_device_config.device_credentials.secret_key,
                "Content-Type": "application/json",
            },
            data=json.dumps(certificate_data),
            timeout=REQUEST_TIME_OUT_SECONDS,
        )
        response.raise_for_status()

    except requests.exceptions.ConnectionError as exc:
        raise Error(MESSAGES.SERVER_CONNECTION_ERROR_MESSAGE) from exc

    except requests.exceptions.HTTPError as exc:
        if response.status_code == 400:
            raise Error(MESSAGES.INSUFFICIENT_DEVICE_QUOTA_ERROR_MESSAGE) from exc

        if response.status_code == 402:
            raise Error(MESSAGES.FEATURE_NOT_AVAILABLE_ERROR_MESSAGE) from exc

        if response.status_code == 403:
            raise Error(
                MESSAGES.SERVER_FORBIDDEN_ERROR_MESSAGE.format(
                    outpost_device_config.device_credentials.workspace_id
                )
            ) from exc

        if response.status_code != 405:
            raise Error(
                MESSAGES.SERVER_UNKNOWN_ERROR_MESSAGE.format(response.text)
            ) from exc

    certificate_status = response.json()["status"]

    if certificate_status["message"] != "Issued":
        raise Error("Error: Certificate signing request failed. Please try again.")

    certificate_path = DEFAULT_CREDENTIALS_DIR / "device.crt"
    with open(certificate_path, "w", encoding="utf-8") as certificate_file:
        certificate_file.write(certificate_status["pemCertificate"])

    os.chmod(
        certificate_path,
        stat.S_IRUSR  # Owner read
        | stat.S_IWUSR  # Owner write
        | stat.S_IRGRP  # Group read
        | stat.S_IROTH,  # Others read
    )

    validate_certificate(outpost_device_config)

    return (
        f"Certificate signing request successful. "
        f"Device certificate saved to {certificate_path}."
    )


@start_spinner
def get_outpost_device_config_files(
    outpost_device_config: OutpostDeviceConfig, message_queue: queue.Queue
) -> str:
    """Download the Outpost configuration files."""
    message_queue.put("Downloading Outpost configuration files...")

    try:
        response = requests.get(
            (
                f"{CONSTS.NEXUS_ENDPOINT}/"
                f"workspaces/ws_{outpost_device_config.device_credentials.workspace_id}"
                f"/outpost/configurations/{outpost_device_config.configuration.id}"
            ),
            headers={
                "Secret-Key": outpost_device_config.device_credentials.secret_key,
                "Content-Type": "application/json",
            },
            verify=True,
            timeout=(60, 3600),
        )
        response.raise_for_status()

    except requests.exceptions.ConnectionError as exc:
        raise Error(MESSAGES.SERVER_CONNECTION_ERROR_MESSAGE) from exc

    except requests.exceptions.HTTPError as exc:
        if response.status_code == 400:
            raise Error(MESSAGES.INSUFFICIENT_DEVICE_QUOTA_ERROR_MESSAGE) from exc

        if response.status_code == 402:
            raise Error(MESSAGES.FEATURE_NOT_AVAILABLE_ERROR_MESSAGE) from exc

        if response.status_code == 403:
            raise Error(
                MESSAGES.SERVER_FORBIDDEN_ERROR_MESSAGE.format(
                    outpost_device_config.device_credentials.workspace_id
                )
            ) from exc

        if response.status_code != 405:
            raise Error(
                MESSAGES.SERVER_UNKNOWN_ERROR_MESSAGE.format(response.text)
            ) from exc

    outpost_configuration = msgspec.json.decode(
        json.dumps(response.json()), type=OutpostDeviceConfiguration
    )

    try:
        config_dir = Path(outpost_device_config.runtime_dir) / "runtime" / "config"
        config_dir.mkdir(parents=True, exist_ok=True)
    except Exception as exc:
        raise Error(f"Error creating config directory: {str(exc)}") from exc

    config_variables = {
        name: entry.default
        for name, entry in outpost_configuration.spec.config_variables.items()
    }

    for (
        config_filename,
        config_file_contents,
    ) in outpost_configuration.spec.config_data.items():
        config_file_contents = interpolate_config_variables(
            config_file_contents, config_variables
        )

        config_file_path = config_dir / config_filename
        with open(config_file_path, "w", encoding="utf-8") as config_file:
            config_file.write(config_file_contents)

    generate_device_config(outpost_device_config, config_dir)
    generate_metadata_spec(outpost_device_config, config_variables, config_dir)
    generate_device_info(outpost_device_config, config_dir)

    return "Outpost configuration files successfully downloaded."


def generate_device_config(
    outpost_device_config: OutpostDeviceConfig, config_dir: Path
) -> None:
    """Generate the device configuration file."""
    config_file_path = config_dir / "device-config.yml"

    variables = {
        "device_id": outpost_device_config.id,
        "device_name": outpost_device_config.name,
        "workspace_id": outpost_device_config.device_credentials.workspace_id,
    }

    device_config_template_str = msgspec.yaml.encode(DEVICE_CONFIG).decode("utf-8")
    device_config_str = interpolate_config_variables(
        device_config_template_str, variables
    )

    with open(config_file_path, "w", encoding="utf-8") as config_file:
        config_file.write(device_config_str)


@start_spinner
def get_outpost_runtime_files(
    outpost_device_config: OutpostDeviceConfig, message_queue: queue.Queue
) -> None:
    """Download the Outpost runtime files."""
    message_queue.put("Downloading Outpost runtime files...")

    signed_url = outpost_device_config.runtime.link.url

    download_response = requests.get(
        signed_url,
        headers={
            "Secret-Key": outpost_device_config.device_credentials.secret_key,
            "Content-Type": "application/json",
        },
        timeout=REQUEST_TIME_OUT_SECONDS,
        stream=True,
    )

    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        for data in download_response.iter_content(chunk_size=1024):
            temp_file.write(data)

    try:
        Path(outpost_device_config.runtime_dir).mkdir(parents=True, exist_ok=True)
    except Exception as exc:
        raise Error(f"Error creating runtime directory: {str(exc)}") from exc

    with zipfile.ZipFile(temp_file.name, "r") as zip_ref:
        zip_ref.extractall(outpost_device_config.runtime_dir)

    return "Outpost runtime files successfully downloaded."


def generate_metadata_spec(
    outpost_device_config: OutpostDeviceConfig,
    config_variables: Dict[str, Any],
    config_dir: Path,
) -> None:
    """Generate the metadata specification file."""
    metadata_spec = MetadataSpec(
        device_uid=outpost_device_config.uid,
        tags=outpost_device_config.tags,
        metadata_generation=outpost_device_config.metadata_generation,
        revision_id=str(outpost_device_config.runtime.version),
        configuration=outpost_device_config.configuration,
        runtime=outpost_device_config.runtime,
        runtime_variables=config_variables,
    )

    metadata_spec_dict = msgspec.to_builtins(metadata_spec)

    with open(config_dir / "metadata-spec.yml", "w", encoding="utf-8") as config_file:
        yaml.dump(metadata_spec_dict, config_file, indent=4, default_flow_style=False)


def generate_device_info(
    outpost_device_config: OutpostDeviceConfig, config_dir: Path
) -> None:
    """Generate the device information file."""
    device_info_dict = msgspec.to_builtins(outpost_device_config.device_info)

    with open(config_dir / "device-info.yml", "w", encoding="utf-8") as config_file:
        yaml.dump(device_info_dict, config_file, indent=4, default_flow_style=False)


def install_runtime(outpost_device_config: OutpostDeviceConfig) -> None:
    """Install the Outpost runtime on the device."""
    with subprocess.Popen(
        [
            "sudo",
            "bash",
            f"{outpost_device_config.runtime_dir}/runtime/scripts/setup.sh",
            "--runtime-version",
            str(outpost_device_config.runtime.version),
            "--config-version",
            str(outpost_device_config.configuration.spec.config_version),
            "--working-dir",
            f"{outpost_device_config.runtime_dir}/runtime",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        bufsize=1,
    ) as installation_process:
        for line in installation_process.stdout:
            print(line, end="")
            sys.stdout.flush()

        # Wait for the process to complete and get the return code
        return_code = installation_process.wait()

        if return_code != 0:
            error_output = installation_process.stderr.read()
            raise Error(
                f"Error installing runtime (code {return_code}): {error_output}"
            )

        print("Runtime installation successful.")


@start_spinner
def deregister_device(
    outpost_device_config: OutpostDeviceConfig, message_queue: queue.Queue
):
    """Deregister the current device from the server."""
    message_queue.put("Deregistering device...")
    try:
        response = requests.delete(
            (
                f"{CONSTS.NEXUS_ENDPOINT}/workspaces/ws_"
                f"{outpost_device_config.device_credentials.workspace_id}"
                f"/outpost/devices/{outpost_device_config.id}"
            ),
            headers={
                "Secret-Key": outpost_device_config.device_credentials.secret_key,
                "Content-Type": "application/json",
            },
            timeout=REQUEST_TIME_OUT_SECONDS,
        )
        response.raise_for_status()

    except requests.exceptions.ConnectionError as exc:
        raise Error(MESSAGES.SERVER_CONNECTION_ERROR_MESSAGE) from exc

    except requests.exceptions.HTTPError as exc:
        if response.status_code == 400:
            raise Error(MESSAGES.INSUFFICIENT_DEVICE_QUOTA_ERROR_MESSAGE) from exc

        if response.status_code == 402:
            raise Error(MESSAGES.FEATURE_NOT_AVAILABLE_ERROR_MESSAGE) from exc

        if response.status_code == 403:
            raise Error(
                MESSAGES.SERVER_FORBIDDEN_ERROR_MESSAGE.format(
                    outpost_device_config.device_credentials.workspace_id
                )
            ) from exc

        if response.status_code == 404:
            return "Device not found. It may have already been deregistered."

        if response.status_code != 405:
            raise Error(
                MESSAGES.SERVER_UNKNOWN_ERROR_MESSAGE.format(response.text)
            ) from exc

    if response.json()["deleted"]:
        return "Device deregistration successful."

    raise Error("Error deregistering device. Please try again.")


def uninstall_runtime(outpost_device_config: OutpostDeviceConfig):
    """Uninstall the Outpost runtime from the device."""
    try:
        with subprocess.Popen(
            [
                "sudo",
                "bash",
                f"{outpost_device_config.runtime_dir}/runtime/scripts/uninstall.sh",
                "--yes",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            bufsize=1,
        ) as process:
            # Stream output in real-time
            for line in process.stdout:
                sys.stdout.write(line)
                sys.stdout.flush()

            # Wait for the process to complete and check return code
            return_code = process.wait()
            if return_code != 0:
                raise subprocess.CalledProcessError(return_code, process.args)

        shutil.rmtree(outpost_device_config.runtime_dir)

    except subprocess.CalledProcessError as exc:
        raise Error(
            f"Uninstall script failed with return code: {exc.returncode}"
        ) from exc

    except Exception as exc:
        raise Error(f"Error running uninstall script: {str(exc)}") from exc

    print("Outpost runtime uninstalled successfully.")


def get_certificate_expiry(certificate_path: Path) -> datetime:
    """Read a certificate file and return its expiration date."""

    with open(certificate_path, "rb") as cert_file:
        cert_data = cert_file.read()

    try:
        cert = x509.load_pem_x509_certificate(cert_data, default_backend())

    except Exception as exc:
        raise Error(f"Error parsing current certificate: {exc}") from exc

    # Get expiration date
    expiry_date = cert.not_valid_after_utc

    return expiry_date


def validate_certificate(outpost_device_config: OutpostDeviceConfig) -> None:
    """Check the validity of the device certificate."""
    certificate_path = None
    for path in [
        RUNTIME_CREDENTIALS_DIR / "device.crt",
        DEFAULT_CREDENTIALS_DIR / "device.crt",
    ]:
        if path.exists():
            certificate_path = path
            break

    if not certificate_path:
        raise Error(
            "Error: Device certificate not found. "
            "Please re-run `datature outpost install` to generate a new certificate."
        )

    outpost_device_config.device_status.certificate_expires_at = get_certificate_expiry(
        certificate_path
    )
    if outpost_device_config.device_status.certificate_expires_at and (
        datetime.now(timezone.utc)
        < outpost_device_config.device_status.certificate_expires_at
    ):
        outpost_device_config.device_status.valid_certificate = True
    else:
        outpost_device_config.device_status.valid_certificate = False


def get_outpost_runtime_status(
    outpost_device_config: OutpostDeviceConfig, user: str = "datature"
):
    """Get the status of the Outpost runtime.

    Args:
        outpost_device_config: Outpost device configuration.
        user: The user under which the service is running.
    """
    device_info = get_device(outpost_device_config)
    overall_status = device_info.get("status", {})
    outpost_device_config.device_status.status = DeviceStatusProperties(
        message=overall_status.get("message"),
        last_connected_at=overall_status.get("lastConnectedAt"),
        sample_prediction=overall_status.get("samplePrediction"),
    )

    try:
        updater_status, updater_pid, updater_status_output = get_service_status(
            "datature-outpost-updater", user
        )
    except Exception as e:
        updater_status = "error"
        updater_pid = None
        updater_status_output = f"Error retrieving status: {str(e)}"

    # Check runtime manager service
    try:
        runtime_status, runtime_pid, runtime_status_output = get_service_status(
            "datature-outpost-runtime-manager", user
        )
    except Exception as e:
        runtime_status = "error"
        runtime_pid = None
        runtime_status_output = f"Error retrieving status: {str(e)}"

    outpost_device_config.device_status.updater_status = UpdaterStatus(
        status=updater_status,
        pid=updater_pid,
        status_output=updater_status_output,
    )

    outpost_device_config.device_status.runtime_status = RuntimeStatus(
        status=runtime_status,
        pid=runtime_pid,
        status_output=runtime_status_output,
    )

    outpost_device_config.write_to_file()


def get_service_status(
    service_name: str,
    user: str = "datature",
) -> Tuple[str, Optional[int], str]:
    """Get the status of a systemd service for a specific user.

    Args:
        service_name: The name of the service to check.
        user: The user under which the service is running.

    Returns:
        Tuple of (active_state, pid, full_output)
    """
    cmd = [
        "sudo",
        "-u",
        user,
        "bash",
        "-c",
        f"XDG_RUNTIME_DIR=/run/user/$(id -u {user}) systemctl --user status {service_name}",
    ]

    try:
        # Run the command and capture output
        result = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            check=False,  # Don't raise exception on non-zero exit code
        )

        # Parse the output to extract status information
        return parse_systemctl_output(result.stdout)

    except subprocess.SubprocessError as exc:
        raise SystemError(
            f"Error running systemctl as user {user}: {str(exc)}"
        ) from exc


def parse_systemctl_output(output: str) -> Tuple[str, Optional[int], str]:
    """
    Parse systemctl status output to extract active state and PID.

    Args:
        output: The output from systemctl status command

    Returns:
        Tuple of (active_state, pid, full_output)
    """
    active_state = "unknown"
    pid = None

    # Process output line by line
    for line in output.splitlines():
        line = line.strip()

        # Extract active state
        if "Active:" in line:
            parts = line.split(":", 1)[1].strip()
            active_state = parts.split()[0].lower()

        # Extract PID
        if "Main PID:" in line:
            pid_part = line.split(":", 1)[1].strip().split()[0]
            try:
                pid = int(pid_part)
            except ValueError:
                pass

    return active_state, pid, output


def format_outpost_runtime_status(
    outpost_device_config: OutpostDeviceConfig, current_device_info: Dict[str, Any]
) -> None:
    """Format and print the Outpost runtime status.

    Args:
        outpost_device_config: Outpost device configuration.
        current_device_info: Current device information.
    """
    overall_status = outpost_device_config.device_status.status
    updater_status = outpost_device_config.device_status.updater_status
    runtime_status = outpost_device_config.device_status.runtime_status

    certificate_expires_in = (
        outpost_device_config.device_status.certificate_expires_at
        - datetime.now(timezone.utc)
    ).days

    is_certificate_valid = (
        outpost_device_config.device_status.valid_certificate
        and certificate_expires_in > 0
    )

    if certificate_expires_in < 0:
        certificate_status = "Expired"
    elif certificate_expires_in < 7:
        certificate_status = "Expiring soon, please renew"
    elif is_certificate_valid:
        certificate_status = "Valid"
    else:
        certificate_status = "Invalid"

    cpu_cores_used = round(current_device_info["cpu_cores_used"], 2)
    cpu_memory_used_gib = round(current_device_info["cpu_memory_used_bytes"] / 1e9, 2)
    disk_space_used_gib = round(current_device_info["disk_space_used_bytes"] / 1e9, 2)
    cameras_in_use = current_device_info["cameras_in_use"]

    print(f"+{'-' * 63}+")
    print(f"| {'NAME':<30} {'ID':<30} |")
    print(
        f"| {outpost_device_config.name:<30} "
        f"{outpost_device_config.id.strip('outpostdevice_'):<30} |"
    )
    print(f"|{'=' * 63}|")
    print(f"| {'CERTIFICATE':<30} {'EXPIRES IN':<30} |")
    print(f"| {certificate_status:<30} {certificate_expires_in:<5} {'days':<24} |")
    print(f"|{'-' * 63}|")
    print(f"| {'Status':<30} {overall_status.message:<30} |")
    print(f"| {'+' * 61} |")
    print(f"| {'SERVICE':<30} {'PID':<10} {'STATUS':<19} |")
    print(f"| {'Updater':<30} {updater_status.pid:<10} {updater_status.status:<19} |")
    print(
        f"| {'Runtime Manager':<30} {runtime_status.pid:<10} {runtime_status.status:<19} |"
    )
    print(f"|{'-' * 63}|")
    print(f"| {'RESOURCE UTILIZATION':<61} |")
    print(f"| {'CPU Cores':<30} {cpu_cores_used:<30} |")
    print(f"| {'CPU Memory':<30} {cpu_memory_used_gib:<10} {'GiB':<19} |")
    print(f"| {'Disk Space':<30} {disk_space_used_gib:<10} {'GiB':<19} |")
    print(
        f"| {'Camera IDs':<30} {', '.join(cameras_in_use) if cameras_in_use else str(None):<30} |"
    )
    print(f"+{'-' * 63}+")

    if overall_status.message == "Archived":
        print(
            "\nThis device has already been archived. To uninstall the runtime, please run the following command: "
            "`datature outpost uninstall`"
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   base_loader.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Base class for model loaders.
"""

# pylint: disable=R0902,W0718

import logging
import os
import queue
import time
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Tuple

import requests

from datature.nexus.cli.deploy.local.consts import DEFAULT_MODEL_DIR
from datature.nexus.cli.deploy.local.utils.common import start_spinner
from datature.nexus.cli.deploy.local.utils.inference import ModelExtension
from datature.nexus.cli.deploy.local.utils.management_api import ManagementAPIClient
from datature.nexus.cli.deploy.local.utils.model import Model
from datature.nexus.error import Error
from datature.nexus.models import LocalArtifact

logger = logging.getLogger("datature-nexus")


class BaseLoader(ABC):
    """Base class for model loaders."""

    _artifact_id: str
    _model_format: str
    _execution_provider: str
    _management_api_client: ManagementAPIClient
    _model_dir: Path
    _model_path: str
    _label_map_path: str
    _model: Model

    def __init__(
        self,
        artifact_id: str,
        execution_provider: str,
        model_format: str,
        management_api_client: ManagementAPIClient,
    ) -> None:
        """Initialize loader."""
        self._artifact_id = artifact_id
        self._execution_provider = execution_provider
        self._model_format = model_format
        self._management_api_client = management_api_client

        self._model_dir = DEFAULT_MODEL_DIR / self._artifact_id
        self._model_path = ""
        self._label_map_path = ""

        self._get_model()

    @abstractmethod
    def load(self, message_queue: queue.Queue = queue.Queue()):
        """Load format-specific model.

        Raises:
            ValueError: Abstract function needs to be implemented.
        """
        raise ValueError("Abstract function needs to be implemented")

    def _get_model(self) -> None:
        """Fetch model from Datature Nexus or local directory.

        Raises:
            InvalidModelException: Error loading model.
        """
        if not os.path.isdir(self._model_dir):
            self._get_model_from_api()

        else:
            try:
                self._get_model_from_local()
            except Exception as exc:
                print(f"Error fetching {self.model_format} model locally: {exc}")
                print(
                    f"Attempting to fetch {self.model_format} model from Datature Nexus..."
                )
                self._get_model_from_api()

    def _get_model_from_api(self) -> None:
        """Fetch model from Datature Nexus."""
        try:
            exported_model = self._export_model()

            local_model = self._download_model(exported_model)

            self._model_path = os.path.join(
                local_model.download_path, local_model.model_filename
            )

            if not os.path.exists(self._model_path):
                raise FileNotFoundError

        except FileNotFoundError as exc:
            raise Error(f"Model file path not found: {self._model_path}!") from exc

        except Exception as exc:
            raise Error(f"Error fetching {self.model_format} model: {exc}") from exc

        try:
            self._label_map_path = os.path.join(
                local_model.download_path, local_model.label_filename
            )

            if not os.path.exists(self._label_map_path):
                raise FileNotFoundError

        except FileNotFoundError as exc:
            raise Error(
                f"Label map file path not found: {self._label_map_path}!"
            ) from exc

        except OSError as exc:
            raise Error("Error getting label map file path") from exc

    def _get_model_from_local(self) -> None:
        """Fetch model from local directory."""
        model_filename = next(
            (
                f
                for f in os.listdir(self._model_dir)
                if any(f.endswith(ext.value) for ext in ModelExtension)
            ),
            None,
        )

        if model_filename is None:
            raise FileNotFoundError(
                f"Model file could not be detected in {self._model_dir}"
            )

        try:
            self._model_path = os.path.join(self._model_dir, model_filename)

            if not os.path.exists(self._model_path):
                raise FileNotFoundError(
                    f"Model file path not found: {self._model_path}"
                )

        except FileNotFoundError as exc:
            raise Error(f"Model file path not found: {self._model_path}") from exc

        except OSError as exc:
            raise Error("Error getting model file path") from exc

        except Exception as exc:
            raise Error(f"Error fetching {self.model_format} model: {exc}") from exc

        try:
            label_map_filename = next(
                (
                    file_name
                    for file_name in os.listdir(self._model_dir)
                    if (file_name.endswith(".pbtxt") or file_name.endswith(".txt"))
                    and "label" in file_name
                ),
                None,
            )

            if label_map_filename is None:
                raise FileNotFoundError

            self._label_map_path = os.path.join(self._model_dir, label_map_filename)

            if not os.path.exists(self._label_map_path):
                raise FileNotFoundError

        except FileNotFoundError as exc:
            raise Error(
                f"Label map file path not found: {self._label_map_path}"
            ) from exc

        except OSError as exc:
            raise Error("Error getting label map file path") from exc

    @start_spinner
    def _export_model(self, message_queue: queue.Queue = queue.Queue()):
        """Export model from Datature Nexus."""
        message_queue.put("Exporting model...")

        while True:
            try:
                exports = (
                    self._management_api_client.project.artifacts.list_exported_models(
                        self._artifact_id
                    )
                )

            except requests.exceptions.HTTPError as exc:
                raise Error(
                    f"Error finding model artifact {self._artifact_id}: {exc.response.text}"
                ) from exc

            exported_model = next(
                (model for model in exports if model.format == self._model_format),
                None,
            )

            if exported_model and exported_model.status == "Error":
                raise Error(
                    "Error exporting model. Please try again later or contact"
                    "support at support@datature.io."
                )

            if not exported_model or exported_model.status != "Finished":
                try:
                    exported_model = (
                        self._management_api_client.project.artifacts.create_export(
                            self._artifact_id,
                            export_options={"format": self._model_format},
                        )
                    )

                except requests.exceptions.HTTPError as exc:
                    # Wait and retry if the model export is already in progress
                    if exc.response.status_code == 409:
                        time.sleep(5)
                        continue

                    raise Error(f"Error exporting model: {exc.response.text}") from exc

            break

        return exported_model, "Model exported successfully"

    @start_spinner
    def _download_model(
        self, exported_model, message_queue: queue.Queue = queue.Queue()
    ) -> Tuple[LocalArtifact, str]:
        """Download model from Datature Nexus."""
        message_queue.put("Downloading model...")

        try:
            downloaded_model = (
                self._management_api_client.project.artifacts.download_exported_model(
                    exported_model["id"], self._model_dir
                )
            )

        except requests.exceptions.HTTPError as exc:
            raise Error(
                f"Server error when downloading model: {exc.response.text}"
            ) from exc

        except Exception as exc:
            raise Error(f"Error downloading model: {exc}") from exc

        return downloaded_model, "Model downloaded successfully"

    @property
    def model_format(self) -> str:
        """Get model format."""
        return self._model_format

    @model_format.setter
    def model_format(self, model_format) -> None:
        """Set model format."""
        self._model_format = model_format

    @property
    def model(self) -> Model:
        """Get loaded model."""
        return self._model

    @model.setter
    def model(self, model) -> None:
        """Set loaded model."""
        self._model = model



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   client.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   SDK Client module
"""
# pylint: disable=C0103

import copy
import os
from abc import ABC
from typing import List, Optional, Union

from datature.nexus import models
from datature.nexus.api.project import Project
from datature.nexus.api.types import ArtifactFilters
from datature.nexus.client_context import ClientContext


class Client(ABC):
    """Initialize the Datature client with a secret key and an API endpoint.

    Parameters:
        secret_key (str): The secret key used for authentication with the Datature API.
        endpoint (str): The base URL for the Datature API.
            Defaults to the value of the DATATURE_API_BASE_URL environment variable,
            or "https://api.datature.io" if the environment variable is not set.
    """

    _context: ClientContext

    def __init__(
        self,
        secret_key: Optional[str] = None,
        mtls_certificate_file_path: Optional[str] = None,
        mtls_private_key_file_path: Optional[str] = None,
        endpoint: str = os.getenv("DATATURE_API_BASE_URL", "https://api.datature.io"),
    ):
        """Initialize"""
        self._context = ClientContext(
            secret_key, mtls_certificate_file_path, mtls_private_key_file_path, endpoint
        )

    def get_info(self) -> models.Workspace:
        """Retrieve this workspace.

        :return: A msgspec struct containing the workspace information with the following structure:

            .. code-block:: python

                Workspace(
                    id='ws_1c8aab980f174b0296c7e35e88665b13',
                    name="Raighne's Workspace",
                    owner='user_6323fea23e292439f31c58cd',
                    tier='Developer',
                    create_date=1701927649302
                )

        :example:

            .. code-block:: python

                from datature.nexus import Client

                client = Client("5aa41e8ba........")
                client.get_info()
        """
        return self._context.requester.GET("/workspace", response_type=models.Workspace)

    def list_projects(self) -> List[Project]:
        """List all projects under this workspace.

        :return: A List of project class instances.
        :example:

            .. code-block:: python

                from datature.nexus import Client

                projects = Client("5aa41e8ba........").list_projects()

                for project in projects:
                    print(project.get_info())
        """
        projects = self._context.requester.GET(
            "/workspace/projects", response_type=models.Projects
        )

        projects_res = []
        for project in projects:
            project_context = copy.deepcopy(self._context)
            project_context.project_id = project.id
            projects_res.append(Project(project_context))

        return projects_res

    def get_project(self, project_id: str) -> Project:
        """Return a project object with a separate context.

        :param project_id: The project ID.
        :return: A class instance of the project.
        :example:

            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")
        """
        assert isinstance(project_id, str)

        if not project_id.startswith("proj_"):
            project_id = f"proj_{project_id}"

        project_context = copy.deepcopy(self._context)
        project_context.project_id = project_id

        return Project(project_context)

    def list_artifacts(
        self,
        filters: Union[ArtifactFilters, dict] = None,
    ) -> models.Artifacts:
        """Lists all artifacts in the workspace.

        :return: A list of msgspec structs containing
            the artifact metadata with the following structure:

            .. code-block:: python

                [Artifact(
                    id='artifact_65ae274540259e2a07533532',
                    project_id='proj_ecfb4589d153999ba9ce01f54568fa72',
                    run_id='run_4a5d406d-464d-470c-bd7d-e92456621ad3',
                    flow_title='Test workflow',
                    artifact_name='ckpt-20',
                    create_date=1705912133684,
                    is_training=False,
                    step=5000,
                    is_deployed=False,
                    metric=ArtifactMetric(
                        total_loss=0.32356,
                        classification_loss=0.012036,
                        localization_loss=0.010706,
                        regularization_loss=0.0
                    ),
                    model_name='fasterrcnn-inceptionv2-1024x1024',
                    export_options=[
                        ArtifactExportOptions(
                            format='ONNX',
                            optimizations=ArtifactExportOptimizations(
                                quantization=['float32', 'float16']
                            ),
                            default_optimizations=ArtifactExportOptimizations(
                                quantization='float32'
                            )
                        ),
                        ArtifactExportOptions(
                            format='TensorFlow',
                            optimizations=ArtifactExportOptimizations(
                                quantization=['float32']
                            ),
                            default_optimizations=ArtifactExportOptimizations(
                                quantization='float32'
                            )
                        ),
                    ],
                    exports=None
                )]

        :example:
            .. code-block:: python

                from datature.nexus import Client

                workspace_artifacts = Client("5aa41e8ba........").list_artifacts()
        """
        assert isinstance(filters, (ArtifactFilters, dict, type(None)))

        if isinstance(filters, dict):
            filters = ArtifactFilters(**filters)
        if filters is None:
            filters = ArtifactFilters()

        workspace_id = self.get_info().id

        return self._context.requester.GET(
            f"/workspaces/{workspace_id}/artifacts",
            query={**filters.to_json()},
            response_type=models.Artifacts,
        )
